##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


from sensiml.method_calls.selectorcall import SelectorCall
from sensiml.datamanager.pipeline import PipelineStep


class SelectorCallSet(PipelineStep):
    """The base class for a collection of selector calls"""

    def __init__(self, name=''):
        super(SelectorCallSet, self).__init__(
            name=name, step_type='SelectorCallSet')
        self._selectors = []
        self._input_data = ''
        self._label_column = None
        self._number_of_features = 10
        self._feature_table = ''
        self._cost_function = ''
        self._remove_columns = []
        self._passthrough_columns = None
        self._refinement = {}

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def input_data(self):
        return self._input_data

    @input_data.setter
    def input_data(self, value):
        self._input_data = value

    @property
    def label_column(self):
        return self._label_column

    @label_column.setter
    def label_column(self, value):
        self._label_column = value

    @property
    def number_of_features(self):
        return self._number_of_features

    @number_of_features.setter
    def number_of_features(self, value):
        self._number_of_features = value

    @property
    def feature_table(self):
        return self._feature_table

    @feature_table.setter
    def feature_table(self, value):
        self._feature_table = value

    @property
    def cost_function(self):
        return self._cost_function

    @cost_function.setter
    def cost_function(self, value):
        self._cost_function = value

    @property
    def refinement(self):
        return self._refinement

    @refinement.setter
    def refinement(self, value):
        self._refinement = value

    @property
    def selectors(self):
        return self._selectors

    def add_selector_call(self, *selectors):
        """Adds one or more selector calls to the collection.

            Args:
                selectors (SelectorCall or list[SelectorCall]): object(s) to append
        """
        for selector in selectors:
            self._selectors.append(selector)

    def remove_selector_call(self, *selectors):
        """Removes one or more selector call from the collection.

            Args:
                selectors (SelectorCall or list[SelectorCall]): object(s) to remove
        """
        for selector in selectors:
            self._selectors = [f for f in self._selectors if f != selector]
        # Note:  We should probably re-populate inputs and outputs here

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, value):
        self._outputs = value

    @property
    def passthrough_columns(self):
        return self._passthrough_columns

    @passthrough_columns.setter
    def passthrough_columns(self, value):
        self._passthrough_columns = value

    @property
    def remove_columns(self):
        return self._remove_columns

    @remove_columns.setter
    def remove_columns(self, value):
        self._remove_columns = value

    def _to_list(self):
        gencalls = []
        for item in self._selectors:
            gencalls.append(item._to_dict())
        return gencalls

    def _to_dict(self):
        selcalls_set = []
        set_dict = {}
        for item in self._selectors:
            selcalls_set.append(item._to_dict())
        set_dict['type'] = 'selectorset'
        set_dict['name'] = getattr(self, '_name')
        set_dict['set'] = selcalls_set
        set_dict['inputs'] = {}
        set_dict['inputs']['remove_columns'] = getattr(self, '_remove_columns')
        set_dict['inputs']['passthrough_columns'] = getattr(
            self, '_passthrough_columns')
        set_dict['inputs']['input_data'] = getattr(self, '_input_data')
        set_dict['inputs']['label_column'] = getattr(self, '_label_column')
        set_dict['inputs']['number_of_features'] = getattr(
            self, '_number_of_features')
        set_dict['inputs']['feature_table'] = getattr(self, '_feature_table')
        if self._cost_function == '':
            self._cost_function = 'sum'
        set_dict['inputs']['cost_function'] = getattr(self, '_cost_function')
        set_dict['outputs'] = self._outputs
        set_dict['refinement'] = getattr(self, '_refinement')
        return set_dict



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


class FunctionCalls(object):
    """Collection of function calls"""

    def __init__(self):
        self._name = ''
        self._function_calls = []

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def function_calls(self):
        return self._function_calls

    def add_function_call(self, function_call):
        """Adds a function call to the collection.

            Args:
                function_call (FunctionCall): object to append
        """
        self._function_calls.append(function_call)

    def remove_function_call(self, function_call):
        """Removes a function call from the collection.

            Args:
                function_call (FunctionCall): object to remove
        """
        self._function_calls = [
            f for f in self._function_calls if f != function_call]

    def _to_list(self):
        fcalls = []
        for item in self.function_calls:
            fcalls.append(item._to_dict())
        return fcalls

        #fcalls['name'] = self.name
        #fcalls['type'] = 'transform'
        #query_dict['outputs'] = [self.outputs]



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################

try:
    from http.cookiejar import CookiePolicy
except ImportError:
    from cookielib import CookiePolicy
import os
import re
import json as JSON
import requests_oauthlib
import requests
from requests import Response
from oauthlib.oauth2 import LegacyApplicationClient, InvalidGrantError, MissingTokenError, TokenExpiredError
from requests_oauthlib import OAuth2Session, OAuth2
from getpass import getpass
from appdirs import user_config_dir
import errno
from uuid import uuid4
from time import sleep
import logging
import warnings
from functools import partial
from .connection_config import ConnectionConfig
from six.moves import input

try:
    import configparser as ConfigParser
except:
    import ConfigParser
from .errors import *

logger = logging.getLogger(__name__)

# Get configuration directory
config_dir = user_config_dir(__name__.split('.')[0], False)

cid = "f7J0QNKkKcIkHMpk5lvq0wGZtBEpw1YcJ8Bea9Pv"
csid = "4OxQzkEnVc6OmTMKHgi77o8uFhEDz1QCuyKsNFm2Jo84r0TlGBXli0NrZ5uHTI769P3UurvDVB86awZkrEikCUJ7RM333BTS5oajO32BRwsNkPcOzA3JlBkPblvKM0l0"


class Warnings:
    insecure_warning = "You have chosen to disable certificate verification. Your connection is not secure!"


class Connection(object):
    """Connection class"""

    def __init__(self, server=None, username=None, password=None, **kwargs):
        """Create Connection object

        Args:
            server (str, optional): Server URL or name in config file.
            username (str, optional)
            password (str, optional)
            **kwargs (optional): Mostly used when passing `server` as URL. See below.

        Keyword Args:
            insecure (bool): Whether or not to require valid SSL/TLS in requests
            path (str, optional): When specifiying a server name instead of a URL, the path of the config file.
            server_name (str, optional): user-readable name of server. defaults to the server domain.
            url (str, optional): Alternative name for server argument.
            auth_url (str, optional): URL of the oauth server. Defaults to server url.
        """
        # Check url parameter to allow direct passing of ConnectionConfig instances

        server = server or kwargs.get('url')
        if re.match(r'^https?://', server):
            self.server_url = re.sub(r'([^/])$', r'\1/', server)
            self.server_name = re.sub(r'^https?://([A-Za-z0-9.\-_]+).*$',
                                      r'\1', server)
            self.auth_url = kwargs.pop('auth_url', self.server_url)
            self.client_id = cid
            self.client_secret = csid
            self.username = username
        else:
            path = kwargs.pop('path', 'connect.cfg')
            config = ConnectionConfig(server)
            try:
                config.load_from_file(path)
            except ConfigParser.NoSectionError:
                choice = input(
                    'No configuration found, would you like to create one? (y)/n ') or 'y'
                if choice.lower() in ['y', 'yes', 't', 'true']:
                    config.create_config_object(path)
                else:
                    raise ConnectionAbortedError()

            self.server_name = config.server
            self.server_url = config.url
            self.auth_url = config.auth_url
            self.client_id = cid
            self.client_secret = csid
            self.username = username

        self.token_url = self.auth_url.rstrip('/') + '/token/'
        self.max_retries = kwargs.pop('max_retries', 3)

        self._token_filename = os.path.join(
            config_dir, '{}_token.json'.format(self.server_name))

        try:
            self._token_last_updated = os.path.getmtime(self._token_filename)
        except OSError as e:
            if e.errno == errno.ENOENT:
                self._token_last_updated = None
            else:
                raise

        # Disable SSL check for localhost or when insecure=True
        self.insecure = kwargs.pop('insecure', False)
        if self.insecure:
            logger.error(Warnings.insecure_warning)
        if re.match(r'^https?://localhost', self.server_url) or self.insecure:
            requests.packages.urllib3.disable_warnings()
            os.environ['OAUTHLIB_INSECURE_TRANSPORT'] = '1'
            self._verify = False
        else:
            os.environ['OAUTHLIB_INSECURE_TRANSPORT'] = ''
            self._verify = True

        self._login(password)

    def _login(self, password, attempt=0):
        if attempt >= self.max_retries:
            raise LoginAttemptsExceededError()

        self.session = OAuth2Session(
            client=LegacyApplicationClient(client_id=self.client_id),
            auto_refresh_url=self.token_url,
            token_updater=self._token_updater
        )
        self._set_session_retries()
        self.session.cookies.set_policy(BlockAll())

        if self.username is None or password is None:
            if os.path.exists(self._token_filename):
                with open(self._token_filename, 'r') as f:
                    self.session.token = JSON.load(f)
            else:
                self.username = input('Enter username: ')
                password = getpass('Enter password: ')

        if not self.session.token:
            try:
                self.session.fetch_token(
                    token_url=self.token_url,
                    username=self.username,
                    password=password,
                    # UAA endpoints always require id/secret as basic auth
                    auth=(self.client_id, self.client_secret),
                    verify=self._verify)
            except (InvalidGrantError, MissingTokenError) as e:
                if isinstance(e, InvalidGrantError):
                    print('Invalid log in credentials, please try again.')
                else:
                    print(
                        'No token received: The remote server did not return a valid authorization response.')
                if os.path.exists(self._token_filename):
                    os.remove(self._token_filename)
                password = None
                self.username = None
                return self._login(password, attempt=attempt + 1)
            self._save_token(self.session.token)
        logger.debug('Login completed')

    def request(self, method, path, json=None, headers=None, **kwargs):
        """Sends request to server with authentication

        Args:
            method (str): HTTP Method.
            path (str): Path from server root to request.
            json (dict, optional): dictionary of information to be submitted as json
            headers (dict, optional): http headers to be sent
            **kwargs: Additional keyword arguments passed to underlying request library.

        Keyword Args:
            data (dict): Data to upload as application/x-www-form-urlencoded or multipart/form-data.
                Do not use in conjunction with json.
            paged (boolean): Return [PagedRequest] iterator instead of response

        Returns:
            Response: Response object from requests library with the result of the request
        """

        attempt = kwargs.pop('attempt', 0)
        paged = kwargs.pop('paged', False)
        verify = kwargs.pop('verify', self._verify)
        url = self.server_url.strip('/') + '/' + path.lstrip('/')
        if self._token_last_updated != os.path.getmtime(self._token_filename):
            with open(self._token_filename, 'r') as f:
                self.session.token = JSON.load(f)
        while attempt < self.max_retries:
            if paged:
                return PagedRequest(self.session, method, url, json=json, headers=headers, verify=verify,
                                    client_id=self.client_id, client_secret=self.client_secret, **kwargs)
            else:
                response = self.session.request(method, url, json=json, headers=headers, verify=verify,
                                                client_id=self.client_id, client_secret=self.client_secret, **kwargs)
            # temp workaround for bug in django-oauth-toolkit
            if response.status_code not in [401]:
                return response
            attempt += 1
            try:
                # If normal autorefresh fails we force it
                token = self.session.refresh_token(
                    self.token_url,
                    auth=(self.client_id, self.client_secret),
                    verify=self._verify
                )
                self._save_token(token)
            except (TokenExpiredError, InvalidGrantError) as e:
                print('Refresh Token expired, please re-enter your login information.')
                try:
                    os.unlink(
                        os.path.join(config_dir, '{}_token.json'.format(
                            self.server_name)))
                except OSError as e:
                    if e.errno != errno.ENOENT:
                        raise
                self._login(None)

        return response

    def paged_request(self, method, path, json=None, headers=None, limit=1000, **kwargs):
        """Returns an iterator that will request a new page each call.

        The result returned by the iterator will be the json decoded data of the
        'results' property of the response. Function args are the same as [request()]

        Args:
            method
            path
            json
            headers
            limit (Optional[int]): Number of entries per page, defaults to 1000.
                Note: this will override the 'limit' param specified in 'params'
                request kwarg.
            **kwargs: passed to request method
        """
        params = kwargs.pop('params', {})
        params['limit'] = limit
        return self.request(method, path, json, headers, params=params, paged=True, **kwargs)

    def file_request(self, path, file_path, data, read_type='rb', method='post', **kwargs):
        """Loads a file and formats metadata for specific endpoints and makes server request

        Args:
            path (str): Path to API endpoint
            file_path (str): Local path to file
            data (dict): Data to send to server along with request. May be sent in nonstandard headers instead of request body for some endpoints.
            method (str, optional): HTTP method, defaults to POST
            read_type (str, optional): Mode to open the file. defaults to binary read ('rb')

        Returns:
            Response
        """
        with open(file_path, read_type) as f:

            # TODO: Make this send params as body instead of headers.
            if path.find('segmenter/') >= 0:
                headers = {
                    'Name': data['name'],
                    'Type': data['specific_type'],
                    'Functioninfile': data['functioninfile']
                }
                request = partial(self.request, method, path,
                                  data=f, headers=headers)

            # TODO: Make this send params as body instead of headers.
            elif path.find('transform/') >= 0:
                headers = {
                    'Name': data['name'],
                    'Type': data['type'],
                    'Functioninfile': data['functioninfile'],
                    'Subtype': data['subtype'],
                    'HasCVersion': data['has_c_version'],
                    'CFileName': data['c_file_name'],
                }
                request = partial(self.request, method, path,
                                  data=f, headers=headers)

            elif path.find('capture/') >= 0:
                request = partial(
                    self.request,
                    method,
                    path,
                    files={'file': f},
                    data={
                        'name': data['name'],
                        'asynchronous': data['asynchronous']
                    })

            elif path.find('eventfile/') >= 0:
                request = partial(
                    self.request,
                    method,
                    path,
                    files={'data_file': f},
                    data={'name': data['name']})

            elif path.find('featurefile/') >= 0:
                request = partial(
                    self.request,
                    method,
                    path,
                    files={'data_file': f},
                    data={'name': data['name']})
            else:
                request = partial(self.request, method, path,
                                  data=data, files={'file': f})

            # call method with previously applied arguments
            return request(**kwargs)

    @staticmethod
    def logout(name):
        """Deletes saved session data for name"""
        try:
            os.unlink(os.path.join(config_dir, '{}_token.json'.format(name)))
        except OSError as e:
            if e.errno == errno.ENOENT:
                print('No session saved for {}.'.format(name))
                return
            else:
                raise
        print('Logged out.')

    def _set_session_retries(self):
        """set the number of retries a session has"""
        self.session.mount(
            "http://",
            requests.adapters.HTTPAdapter(max_retries=self.max_retries))
        self.session.mount(
            "https://",
            requests.adapters.HTTPAdapter(max_retries=self.max_retries))

    def _token_updater(self, token, save=True):
        """a simple function to save our token on refresh

        Args:
            token (dict): a token object for oauth2
        """
        self.session.token = token
        if save:
            self._save_token(token)

    def _save_token(self, token):
        if not os.path.exists(config_dir):
            try:
                os.makedirs(config_dir)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise
        with open(self._token_filename, 'w') as f:
            # f.write(json.dumps(token))
            JSON.dump(token, f)
            self._token_last_updated = os.path.getmtime(f.name)


# With ouath we don't want to send cookies, this will overwite and block session cookies
class BlockAll(CookiePolicy):
    return_ok = set_ok = domain_return_ok = path_return_ok = lambda self, * \
        args, **kwargs: False
    netscape = True
    rfc2965 = hide_cookie2 = False


class PagedRequest(object):
    def __init__(self, session, method, path, json=None, headers=None, **kwargs):
        self.session = session
        self._method = method
        self._path = path
        self._json = json
        self._headers = headers
        self._kwargs = kwargs

        self.response = None

    def __iter__(self):
        return self

    def next(self):
        """
        Returns:
            (dict) content of the results property in the response
        """
        if not self.response:
            self.response = self.session.request(
                self._method, self._path, self._json, self._headers, **self._kwargs)
        else:
            json = self.response.json()
            path = json['next']
            if not path:
                raise StopIteration
            self.response = self.session.request(
                self._method, path, self._json, self._headers, **self._kwargs)
        self.response.raise_for_status()
        return self.response.json()['results']



from sensiml.widgets.auto_sense import AutoSenseWidget
from sensiml.widgets.download_widget import DownloadWidget
from sensiml.widgets.query_widget import QueryWidget
from sensiml.widgets.flash_widget import FlashWidget
from sensiml.widgets.model_visualization_widget import ModelVisualizationWidget
from sensiml.widgets.model_explore_widget import ModelExploreWidget
from sensiml.widgets.dashboard_widget import DashBoard


__all__ = ['AutoSenseWidget', 'DownloadWidget', 'QueryWidget', 'FlashWidget',
           "ModelVisualizationWidget", "DashBoard", "ModelExploreWidget"]



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017-18  SensiML Corporation.                                   #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################

from ipywidgets import widgets
from ipywidgets import Layout, Button, HBox, VBox, Accordion, FloatText, Textarea, Dropdown, Label, IntSlider, Checkbox, Text, Button, SelectMultiple, Password
from IPython.display import display
from ipywidgets import IntText
from json import dumps as jdump
from sensiml.widgets.base_widget import BaseWidget
from IPython.display import clear_output
from sensiml.widgets import QueryWidget, AutoSenseWidget, DownloadWidget, FlashWidget, ModelExploreWidget
from IPython.core.display import HTML
try:
    display(HTML("<style>.container { width:90% !important; }</style>"))
except:
    pass
layout = Layout(border='outset',
                width='100%')


class DashBoard(BaseWidget):

    def __init__(self, dsk=None, server='https://sensiml.cloud/', knowledgepack_level='pipeline', **kwargs):
        self._server = server
        self._kwargs = kwargs
        self._dsk = dsk

        if self._dsk:
            self._dsk._auto_sense_ran = False
            self._dsk._query_created = False

        self._create_widget(knowledgepack_level=knowledgepack_level)

    def _set_project(self, option):

        if not isinstance(option['new'], dict):
            return
        if option['new'].get('index', None) is None:
            return

        project_name = option['owner'].options[option['new']['index']]

        if not project_name:
            return

        self._dsk.project = project_name

        pipelines = self._dsk.list_sandboxes()
        if pipelines is not None:
            pipelines = [
                '']+sorted(list(pipelines.Name.values), key=lambda s: s.lower())
        else:
            pipelines = []

        self._widget_pipeline.options = pipelines
        self._widget_pipeline.observe(self._set_pipeline)
        self._refresh_project()

    def _set_pipeline(self, option):

        if not isinstance(option['new'], dict):
            return
        if option['new'].get('index', None) is None:
            return
        if not option['new'].get('index', None):
            return

        pipeline_name = option['owner'].options[option['new']['index']]
        self._dsk.pipeline = pipeline_name

        self._refresh_pipelines()

    def _logout(self, b):
        if self._dsk:
            self._dsk.logout()
            self._dsk = None
            self._clear_cell_output(None)

            self._widget_project.options = []
            self._widget_pipeline.options = []
            self._auto._clear()
            self._download._clear()
            self._explore._clear()
            self._button_connected.description = ''
            self._button_connected.icon = ''
            self._button_connected.style.button_color = '#00b300'
            self._button_connected.disabled = True

    def _clear_cell_output(self, b, return_display=True):
        clear_output(wait=True)

        if return_display:
            return display(self.v)

    def _on_button_clicked(self, b):
        from sensiml import SensiML

        if self._dsk is None:
            try:
                self._dsk = SensiML(server=self._server)

                self._dsk._auto_sense_ran = False
                self._dsk._query_created = False
            except:
                print("Login Failed.")
                return

        self._clear_cell_output(None, return_display=False)

        self._button_connected.description = 'Connected'
        self._button_connected.icon = 'link'
        self._button_connected.style.button_color = '#0075c1'

        self._widget_project.options = [
            '']+sorted(list(self._dsk.list_projects().Name.values), key=lambda s: s.lower())
        self._widget_project.observe(self._set_project)
        self._refresh_login()

    def _add_pipeline(self, b):

        if self._dsk is None:
            print("You must login first.")
            return
        if self._dsk.project is None:
            print("Project Must be set first.")
            return

        if not self._widget_new_pipeline.value:
            return None

        if self._widget_new_pipeline.value not in self._widget_pipeline.options:
            self._dsk.pipeline = self._widget_new_pipeline.value

        self._widget_pipeline.options = self._widget_pipeline.options + \
            (self._widget_new_pipeline.value,)
        self._widget_pipeline.value = self._widget_new_pipeline.value

        self._refresh_pipelines()

    def _create_widget(self, knowledgepack_level='Project'):

        self._button_connected = Button(description="Connected")
        self._button_loggout = Button(description="Logout")
        self._button_clear = Button(
            description="Clear Log", tooltip='Clear Log Text Below Dashboard')
        self._widget_project = Dropdown(description="Project", options=[])
        self._widget_pipeline = Dropdown(description="Pipeline", options=[])
        self._widget_new_pipeline = Text(description="New Pipeline", value='')
        self._button_add_pipeline = Button(description="Add", icon='plus')

        self._button_connected.on_click(self._on_button_clicked)
        self._button_clear.on_click(self._clear_cell_output)
        self._button_add_pipeline.on_click(self._add_pipeline)
        self._button_loggout.on_click(self._logout)

        login = VBox([
            HBox([VBox([self._button_connected,
                        self._button_loggout]),
                  VBox([self._widget_project,
                        self._widget_pipeline]),
                  VBox([HBox([self._widget_new_pipeline, self._button_add_pipeline]),
                        ])
                  ])], layout=layout)

        bottom = VBox([self._button_clear], layout=layout)

        self._query = QueryWidget(None)
        self._auto = AutoSenseWidget(None)
        self._download = DownloadWidget(None, level=knowledgepack_level)
        self._flash = FlashWidget(None)
        self._explore = ModelExploreWidget(None, level=knowledgepack_level)

        accordions = [Accordion(children=[self._query.create_widget()]),
                      Accordion(children=[self._auto.create_widget()]),
                      Accordion(children=[self._explore.create_widget()]),
                      Accordion(children=[self._download.create_widget()]),
                      Accordion(children=[self._flash.create_widget()])

                      ]
        accordions[0].set_title(0, 'Data Exploration')
        accordions[1].set_title(0, 'Model Building')
        accordions[2].set_title(0, 'Explore Models')
        accordions[3].set_title(0, 'Create Knowledge Pack')
        accordions[4].set_title(0, 'Deploy Knowledge Pack')

        for a in accordions:
            a.selected_index = None
        self.v = VBox([login] + accordions + [bottom])
        self.v.layout.border = "outset"

        self._on_button_clicked(None)

        if self._dsk:
            display(self.v)

    def _refresh_login(self):
        self._flash._dsk = self._dsk
        self._flash._refresh()

    def _refresh_project(self):
        self._query._dsk = self._dsk
        self._query._refresh()
        self._auto._clear()
        self._download._clear()
        self._explore._clear()

    def _refresh_pipelines(self):
        self._auto._dsk = self._dsk
        self._auto._refresh()
        self._download._dsk = self._dsk
        self._download._refresh()
        self._explore._dsk = self._dsk
        self._explore._refresh()



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017-18  SensiML Corporation.                                   #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


from ipywidgets import widgets
from ipywidgets import Layout, Button, HBox, VBox, Box, HTML, FloatText, Tab, Textarea, Dropdown, Label, IntSlider, Checkbox, Text, Button, SelectMultiple
from IPython.display import display
from ipywidgets import IntText
from sensiml.widgets.base_widget import BaseWidget, button_handling
from sensiml.base.exceptions import QueryExistsException
from six import text_type
import numpy as np
from IPython.display import display
from bqplot import (
    OrdinalScale, LinearScale, Bars, Lines, Axis, Figure
)
from sensiml.widgets.renderers import WidgetAttributeRenderer


class QueryWidget():

    def __init__(self, dsk):
        self._dsk = dsk
        self._items = {'query_select': ''}

    def _enable_buttons(self):

        self._add_query.disabled = False
        self._update_query.disabled = False
        self._delete_query.disabled = False

    def _disable_buttons(self):

        self._add_query.disabled = True
        self._update_query.disabled = True
        self._delete_query.disabled = True

    @button_handling
    def on_add_query_button_clicked(self, b):

        if self.add_query(create=True, force=False):
            self.query_widget_tab.selected_index = 0

    @button_handling
    def on_update_query_button_clicked(self, b):
        if self._query_select.value:
            self.add_query(force=True)

    @button_handling
    def on_delete_query_button_clicked(self, b):
        if self._dsk is None:
            return
        if self._dsk.project is None:
            return
        if self._query_select.value:
            query = self._dsk.project.queries.get_query_by_name(
                self._query_select.value)
            query.delete()

        self._refresh()

    def add_query(self, create=False, force=False):
        if self._dsk is None:
            return
        if self._dsk.project is None:
            return
        if self._query_select.options == self._query_name_text.value and force == False:
            return

        self.populate_items(populate_create=create)
        query_name = self._items.get('query_name', None)
        segmenter_name = self._items.get('segmenter_name', None)
        label_column = self._items.get('label_column')
        metadata_columns = self._items.get('metadata_columns', None)
        sensor_columns = self._items.get('sensor_columns', None)
        metadata_filter = self._items.get('metadata_filter', '')

        if not segmenter_name:
            self.renderer.render("No segmenter specified")
            return

        # reverse lookup of segmenter id
        segmenters = self._dsk.list_segmenters()
        segmenter_id = int(
            segmenters[segmenters['name'] == self._items['segmenter_name']]['id'].values[0])

        if not query_name:
            self.renderer.render("No query name specified")
            return

        if label_column in metadata_columns:
            self.renderer.render(
                "Label Column cannot be part of metadata columns.")
            return

        if not metadata_columns:
            self.renderer.render("No Metadata Columns Specified")
            return
        if isinstance(metadata_columns, str):
            metadata_columns = metadata_columns.split(',')

        if not sensor_columns:
            self.renderer.render("No Sensor Columns Specified")
            return

        if isinstance(sensor_columns, str):
            sensor_columns = sensor_columns.split(',')

        try:
            self.renderer.render('Creating Query.')
            self.query = self._dsk.create_query(query_name,
                                                columns=sorted(sensor_columns),
                                                metadata_columns=[
                                                    m for m in metadata_columns if m != label_column],
                                                segmenter=segmenter_id,
                                                metadata_filter=metadata_filter,
                                                label_column=label_column,
                                                force=force)
            self.renderer.render('Query Created.')

        except QueryExistsException:
            self.renderer.render(
                'Query Already Exists. Select force to overwrite.')
            return None

        if create:
            self._query_select.options = sorted(
                self._query_select.options + (query_name,), key=lambda s: s.lower())
            self._query_select.value = query_name
            self._query_name_text.value = ''
            self._metadata_filter_text_create.value = ''
            self._segmenter_select_create.value = ''

        else:
            self._query_select.options = sorted(
                self._query_select.options, key=lambda s: s.lower())
            self._query_select.value = query_name
            self.update_statistics(self.query)

        return True

    @button_handling
    def update_statistics(self, query):

        try:
            self.renderer.render("Getting Query Statistics.")
            stats = query.statistics()
            self.renderer.render("")

        except:
            self._graph.marks = []
            self._graph.axes = []
            self._graph.title = ''
            return

        Label = list(stats.sum().index)
        counts = list(stats.sum().values)

        x_ord = OrdinalScale()
        y_lin = LinearScale(min=0)

        ax_x = Axis(scale=x_ord, tick_values=Label)
        ax_y = Axis(scale=y_lin, orientation='vertical', num_ticks=5)

        bar = Bars(x=Label, y=counts, color_mode='group',
                   scales={'x': x_ord, 'y': y_lin})

        self._graph.marks = [bar]
        self._graph.axes = [ax_x, ax_y]
        self._graph.title = ''

    def _refresh(self):
        self._dsk.project.refresh()

        sensor_items = []
        metadata_items = []
        queries = []

        segmenters = self._dsk.list_segmenters()
        if segmenters is None:
            self.renderer.render("No queryable data added to this project")
            segmenters = []
        else:
            segmenters = sorted(
                list(segmenters['name'].values), key=lambda s: s.lower())

        columns = self._dsk.project.columns()
        if columns is None:
            self.renderer.render("No schema associated with this project.")
            columns = []

        for column in sorted(columns):
            sensor_items.append(column)

        metadata_items = self._dsk.project.metadata_columns()
        if metadata_items is None:
            self.renderer.render("No metdata associated with this project")
            metadata_items = []

        label_items = []
        for label in self._dsk.project.label_columns():
            if label != 'SegmentID':
                label_items.append(label)

        query_df = self._dsk.list_queries()
        if query_df is not None:
            for query in query_df['Name'].values:
                queries.append(query)

        self._segmenter_select.options = [''] + segmenters
        self._segmenter_select_create.options = [''] + segmenters
        self._segmenter_select.value = self._segmenter_select.options[0]
        self._label_column_select.options = label_items
        self._label_column_select_create.options = label_items

        self._metadta_items_mselect.options = metadata_items
        self._metadata_items_mselect_create.options = metadata_items
        for name in metadata_items:
            if name.lower() in ['subject', 'subjects']:
                self._metadata_items_mselect_create.value = [name]

        self._sensor_columns_mselector.options = sensor_items
        self._sensor_columns_mselector_create.options = sensor_items
        self._sensor_columns_mselector_create.value = sensor_items
        self._query_select.options = [
            ''] + sorted(list(set(queries)), key=lambda s: s.lower())
        self._query_name_text.value = ''
        self._metadata_filter_text_create.value = ''
        self._metadata_filter_text.value = ''
        self.populate_items()

        return True

    def populate_items(self, populate_create=False):

        self._items['sensor_columns'] = self._sensor_columns_mselector.value
        self._items['metadata_columns'] = self._metadta_items_mselect.value
        self._items['label_column'] = self._label_column_select.value
        self._items['segmenter_name'] = self._segmenter_select.value
        self._items['query_name'] = self._query_select.value
        self._items['query_select'] = self._query_select.value
        self._items['metadata_filter'] = self._metadata_filter_text.value

        if populate_create:
            self._items['sensor_columns'] = self._sensor_columns_mselector_create.value
            self._items['metadata_columns'] = self._metadata_items_mselect_create.value
            self._items['label_column'] = self._label_column_select_create.value
            self._items['segmenter_name'] = self._segmenter_select_create.value
            self._items['query_name'] = self._query_name_text.value
            self._items['metadata_filter'] = self._metadata_filter_text_create.value

    def populate_query(self, change=None):
        self.populate_items()

        self.query = self._dsk.project.queries.get_query_by_name(
            self._items['query_select'])

        if self.query:
            qdict = self.query._to_dict()
            self._qdict = qdict
            seg = self._dsk.list_segmenters()

            self._sensor_columns_mselector.value = qdict['sensor_columns']
            self._metadta_items_mselect.value = qdict['metadata_columns']
            self._label_column_select.value = qdict['label_column']
            self._metadata_filter_text.value = qdict['metadata_filter']

            if list(seg[seg['id'] == qdict['segmenter_id']]['name']):
                self._segmenter_select.value = list(
                    seg[seg['id'] == qdict['segmenter_id']]['name'])[0]
            else:
                self.renderer.render("Error. Query {} is associated with a Segmenter that no longer exists!".format(
                    self._items['query_select']))
                return

            self.populate_items()

            self.update_statistics(self.query)

    def create_widget(self, name=''):
        # refresh the project so that the most up to date values are shown

        self._query_name_text = Text(description='Name')
        self._query_select = Dropdown(description='Name')
        self._segmenter_select = Dropdown(description='Segmenter')
        self._segmenter_select_create = Dropdown(description='Segmenter')
        self._label_column_select = Dropdown(description='Label')
        self._label_column_select_create = Dropdown(description='Label')
        self._metadta_items_mselect = SelectMultiple(description='Metadata')
        self._metadata_items_mselect_create = SelectMultiple(
            description='Metadata')
        self._sensor_columns_mselector = SelectMultiple(description='Sources')
        self._sensor_columns_mselector_create = SelectMultiple(
            description='Sources')
        self._metadata_filter_text = Text(description='Query Filter')
        self._metadata_filter_text_create = Text(description='Query Filter')
        self._add_query = Button(icon='plus', description='Add', tooltip='Create New Query', layout=Layout(
            width='98%', allign_item='center'))
        self._update_query = Button(icon='upload', description='Update',
                                    tooltip='Update Query', layout=Layout(width='98%', allign_item='center'))
        self._delete_query = Button(icon='trash', description='Delete', tooltip='Delete Query', layout=Layout(
            width='98%', allign_item='center'))

        self._add_query.on_click(self.on_add_query_button_clicked)
        self._update_query.on_click(self.on_update_query_button_clicked)
        self._delete_query.on_click(self.on_delete_query_button_clicked)

        self._graph = Figure(marks=[], axes=[], title='',
                             legend_location='bottom-right',
                             background_style={'fill': 'white'},
                             fig_margin={"top": 10, "bottom": 30, "left": 50, "right": 0},)
        self._graph.layout.height = text_type("325px")
        self._graph.layout.width = text_type("560px")

        self._query_select.observe(self.populate_query, 'value')

        self._widget_render_space = HTML()

        self.renderer = WidgetAttributeRenderer(
            self._widget_render_space, 'value')

        self.query_widget_tab = Tab([HBox([VBox([self._query_select,
                                                 self._segmenter_select,
                                                 self._label_column_select,
                                                 self._metadta_items_mselect,
                                                 self._sensor_columns_mselector,
                                                 self._metadata_filter_text,
                                                 HBox(
                                                     [self._delete_query, self._update_query]),
                                                 ]),
                                           self._graph]),
                                     HBox([VBox([self._query_name_text,
                                                 self._segmenter_select_create,
                                                 self._label_column_select_create,
                                                 self._metadata_items_mselect_create,
                                                 self._sensor_columns_mselector_create,
                                                 self._metadata_filter_text_create,
                                                 self._add_query
                                                 ])])
                                     ])

        self.query_widget = VBox([self.query_widget_tab,
                                  self._widget_render_space
                                  ])

        self.query_widget_tab.set_title(0, 'Select/Modify Query')
        self.query_widget_tab.set_title(1, 'Create New Query')

        if self._dsk:
            if self._refresh():
                self.populate_query()

        return self.query_widget



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


import pandas as pd
from os.path import dirname
from os.path import join
import numpy as np


class ClassLabelException(Exception):
    pass


class DataSets():
    def __init__(self):
        pass

    def load_activity_raw(self):
        module_path = dirname(__file__)
        return pd.read_csv(join(module_path, 'data', 'activity_raw.csv'))

    def load_activity_raw_toy(self):
        module_path = dirname(__file__)
        return pd.read_csv(join(module_path, 'data', 'activity_raw_toy.csv'))

    def load_activity_norm(self):
        module_path = dirname(__file__)
        return pd.read_csv(join(module_path, 'data', 'activity_norm.csv'))

    def load_activity_quant(self):
        module_path = dirname(__file__)
        return pd.read_csv(join(module_path, 'data', 'activity_quant.csv'))

    def load_activity_quant_toy(self):
        module_path = dirname(__file__)
        return pd.read_csv(join(module_path, 'data', 'activity_quant_toy.csv'))

    def load_gesture_raw(self):
        module_path = dirname(__file__)
        return pd.read_csv(join(module_path, 'data', 'gesture_raw.csv'))

    def load_toy_data(self):
        module_path = dirname(__file__)
        return pd.read_csv(join(module_path, 'data', 'toy_data.csv'))

    @staticmethod
    def generate_step_data(window_size=100, num_classes=5, num_iterations=2, scale_factor=100, noise_scale=5, class_labels=None, class_values=None):
        """Generate a labeled test data vector.

        Args:
            window_size (int, optional): The size of each segment window.
            num_classes (int, optional): The number of separate classes
            num_iterations (int, optional): The number of iterations to create.
            scale_factor (int, optional): scale the values of the data by this much before adding noise
            noise_scale (int, optional): The range of noise to add (-noise_scale:noise_scale)
            class_labels (None, optional): you can pass specific class labels as an array, otherwise defaults to
             the range(num_classes)

        Returns:
            DataFrame: a test data vector for building and testing models
        """
        class_labels = check_class_label(num_classes, class_labels)
        class_values = check_class_values(num_classes, class_values)
        delta = window_size*num_classes
        signal_length = num_iterations*delta

        y = np.zeros(signal_length)
        y_label = np.zeros(signal_length)

        for j in range(num_iterations):
            y_temp = np.zeros(window_size*num_classes)
            if j % 2 == 0:
                for i in range(num_classes):

                    y_temp[window_size*i:window_size*(i+1)] = class_values[i]
                    y_label[window_size*i+j*delta:window_size *
                            (i+1)+j*delta] = class_labels[i]
            else:
                for i in range(num_classes):
                    y_temp[window_size*i:window_size *
                           (i+1)] = class_values[-1-i]
                    y_label[window_size*i+j*delta:window_size *
                            (i+1)+j*delta] = class_labels[num_classes-1-i]
            y[delta*j:delta*(j+1)] = y_temp*scale_factor

        if noise_scale > 0:
            y += np.array([np.random.choice(range(-noise_scale, noise_scale))
                           for i in range(len(y))])

        return pd.DataFrame({'Data': y.astype(np.int16), 'Label': y_label.astype(int)})

    @staticmethod
    def generate_harmonic_data(window_size=100, num_classes=5, num_iterations=2, scale_factor=100, noise_scale=5, class_labels=None, harmonic_nodes=None):
        """Generate a labeled test data sin vector.

        Args:
            window_size (int, optional): The size of each segment window.
            num_classes (int, optional): The number of separate classes
            num_iterations (int, optional): The number of iterations to create.
            scale_factor (int, optional): scale the values of the data by this much before adding noise
            noise_scale (int, optional): The range of noise to add (-noise_scale:noise_scale)
            class_labels (None, optional): you can pass specific class labels as an array, otherwise defaults to
             the range(num_classes)

        Returns:
            DataFrame: a test data vector for building and testing models
        """

        class_labels = check_class_label(num_classes, class_labels)
        harmonic_nodes = check_class_values(num_classes, harmonic_nodes)

        delta = window_size*num_classes
        signal_length = num_iterations*delta

        y = np.zeros(signal_length)
        x = np.arange(signal_length)

        y_label = np.zeros(signal_length)
        for j in range(num_iterations):
            y_temp = np.zeros(delta)
            x = np.arange(window_size)
            if j % 2 == 0:
                for i in range(num_classes):
                    y_temp[window_size*i:window_size *
                           (i+1)] = get_standing_wave(x, window_size, harmonic_nodes[i])
                    y_label[window_size*i+j*delta:window_size *
                            (i+1)+j*delta] = class_labels[i]
            else:
                for i in range(num_classes):
                    y_temp[window_size*i:window_size *
                           (i+1)] = get_standing_wave(x, window_size, harmonic_nodes[-1-i])
                    y_label[window_size*i+j*delta:window_size *
                            (i+1)+j*delta] = class_labels[num_classes-1-i]
            y[delta*j:delta*(j+1)] = y_temp*scale_factor

        if noise_scale > 0:
            y += np.array([np.random.choice(range(-noise_scale, noise_scale))
                           for i in range(len(y))])

        return pd.DataFrame({'Data': y.astype(np.int16), 'Label': y_label.astype(int)})


def get_standing_wave(x, L, n):
    ld = 2. * L / n
    return np.sin(x*(2*np.pi) / ld)


def check_class_label(num_classes, class_labels=None):
    if class_labels is None:
        class_labels = range(1, num_classes+1)
    elif len(class_labels) != num_classes:
        raise ClassLabelException(
            "The number of classes is not equal to the number of class labels.")

    for i in class_labels:
        if type(i) != type(1):
            raise ClassLabelException("class labels must be integers.")
    if i <= 0:
        raise ClassLabelException("class labels must be greater than zero.")

    return class_labels


def check_class_values(num_classes, class_values=None):
    if class_values is None:
        class_values = range(1, num_classes+1)
    elif len(class_values) != num_classes:
        raise ClassLabelException(
            "The number of classes is not equal to the number of class values.")

    for i in class_values:
        if type(i) != type(1):
            raise ClassLabelException("class values must be type int")
        if i <= 0:
            raise ClassLabelException(
                "class values must be greater than zero.")

    return class_values





def button_handling(func):
    def wrapper(self, *args, **kwargs):
        self._disable_buttons()
        try:
            func(self, *args, **kwargs)
        except Exception as e:
            self._enable_buttons()
            raise e
        self._enable_buttons()

    return wrapper


class BaseWidget():
    def __init__(self, dsk):
        self._dsk = dsk
        self.setup()

    def setup(self):
        pass

    def on_button_clicked(self, b):
        pass

    def create_widget(self):
        pass

    def _enable_buttons(self):
        pass

    def _disable_buttons(self):
        pass



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


class BaseMethodCall(object):
    """The base class for calls to functions.
    Child classes have their own additional properties and overwrite this docstring.
    """

    def __init__(self, name='', function_type=''):
        self._name = name
        self._type = function_type

    def _public_properties(self):
        """Returns the object's public properties, i.e. those that do not start with _"""
        return (name for name in dir(self) if not name.startswith('_'))

    def _to_dict(self):
        prop_dict = {}
        prop_dict['inputs'] = {}
        for prop in self._public_properties():
            if getattr(self, prop) is not None:
                prop_dict['inputs'][prop] = getattr(self, prop)
        prop_dict['name'] = self._name
        return prop_dict



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################

import json
from sensiml.datamanager.sandbox import Sandbox
import sensiml.base.utility as utility


class SandboxExistsError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class Sandboxes():
    """Base class for a collection of sandboxes."""

    def __init__(self, connection, project):
        self._connection = connection
        self._project = project

    def build_sandbox_list(self):
        """Populates the function_list property from the server."""
        sandbox_list = {}

        sandbox_response = self.get_sandboxes()
        for sandbox in sandbox_response:
            sandbox_list[sandbox.name] = sandbox

        return sandbox_list

    def get_or_create_sandbox(self, name):
        """Calls the REST API and gets the sandbox by name, if it doesn't exist insert a new sandbox

            Args:
                name (str): name of the sandbox

            Returns:
                sandbox object
        """
        sandbox = self.get_sandbox_by_name(name)

        if sandbox is None:
            print("Sandbox {} does not exist, creating a new sandbox.".format(name))
            sandbox = self.new_sandbox()
            sandbox.name = name
            sandbox.insert()

        return sandbox

    def create_sandbox(self, name, steps):
        """Creates a sandbox using the given name and steps

            Args:
                name (str): name of the sandbox
                steps (list[dict]): list of dictionaries specifying the pipeline steps

            Returns:
                sandbox object
        """
        if self.get_sandbox_by_name(name) is not None:
            raise SandboxExistsError(
                "sandbox {0} already exists.".format(name))
        else:
            sandbox = self.new_sandbox()
            sandbox.name = name
            sandbox._steps = steps
            sandbox.insert()
            return sandbox

    def get_sandbox_by_name(self, name):
        """Calls the REST API and gets the sandbox by name

            Args:
                name (str): name of the sandbox

            Returns:
                sandbox object
        """
        sandbox_list = self.get_sandboxes()
        for sandbox in sandbox_list:
            if sandbox.name == name:
                return sandbox
        return None

    def new_sandbox(self, dict={}):
        """Creates a new sandbox but does not insert in on the server

            Args:
                dict (optional[dict]): contains name and steps properties of the sandbox

            Returns:
                sandbox object
        """
        sandbox = Sandbox(self._connection, self._project)
        if len(dict) is not 0:
            sandbox.initialize_from_dict(dict)
        return sandbox

    def get_sandboxes(self):
        """Gets all sandboxes from server and turns them into local sandbox objects.

            Returns:
                list[sandbox]
        """
        # Query the server and get the json
        url = 'project/{0}/sandbox/'.format(self._project.uuid)
        response = self._connection.request('get', url,)
        try:
            response_data, err = utility.check_server_response(response)
        except ValueError as e:
            print(response)
        # Populate each sandbox from the server
        sandboxes = []
        if err is False:
            for sandbox_params in response_data:
                sandboxes.append(self.new_sandbox(sandbox_params))
        return sandboxes

    def __str__(self):
        output_string = 'Sandboxes:\n'
        for s in self.get_sandboxes():
            output_string += '    {0}\n'.format(s.name)

        return output_string



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################

from six import string_types
from sensiml.datamanager.pipeline import PipelineStep


class FunctionCall(PipelineStep):
    """The base class for a function call.

    Child classes have their own additional properties and overwrite this docstring."""

    def __init__(self, name, function_type='transform'):
        super(FunctionCall, self).__init__(name=name, step_type='FunctionCall')
        self._type = function_type

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, outputs):
        self._outputs = outputs

    def _check_inputs(self, output_vars=[]):
        """Compares the list of inputs with the given output_vars list

            Args:
                output_vars (list[str]): the current set of output variables

            Returns:
                unmatched_inputs (list[str]): unmatched inputs or an empty list if all inputs are matched
        """
        unmatched_inputs = []
        for param in [getattr(self, input['name']) for input in self._input_contract]:
            if isinstance(param, string_types) and param.find('temp.') > -1 and not param in output_vars:
                unmatched_inputs.append(param)
        return unmatched_inputs

    def _public_properties(self):
        """Returns the object's public properties, i.e. those that do not start with _"""
        return (name for name in dir(self) if not name.startswith('_'))

    def _to_dict(self):
        prop_dict = {}
        prop_dict['name'] = self._name
        prop_dict['type'] = self._type
        prop_dict['feature_table'] = self._feature_table
        prop_dict['inputs'] = {}
        for prop in [input['name'] for input in self._input_contract if not input.get('handle_by_set', False)]:
            if prop != 'outputs':
                prop_dict['inputs'][prop] = getattr(self, prop)
        prop_dict['outputs'] = getattr(self, 'outputs')
        return prop_dict



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017-18  SensiML Corporation.                                   #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


from sensiml.datamanager.pipelineseed import PipelineSeed


class empty():
    pass


param_tabs = ',\n'+' '*4*8


def function_help(func):
    filter_list = ['group_columns', 'label_column', 'ignore_columns', 'passthrough_columns',
                   'classifiers', 'validation_methods']
    params = []
    for item in func.input_contract:
        key = item.get('name')
        value = item.get('default', None)
        key_type = item.get('type', None)
        options = item.get('options', None)
        if value is not None:
            if key_type == 'str':
                params.append('"{}": "{}"'.format(key, value))
            else:
                params.append('"{}": {}'.format(key, value))

            if options is not None and key not in filter_list:
                values = map(lambda x: x['name'], options)
                params[-1] += ', #options: ' + '<{}>'.format(
                    '/'.join(values))
        elif options and key not in filter_list:
            values = map(lambda x: x['name'], options)
            params.append('"{}": <{}>'.format(key, '/'.join(values)))
        elif key_type in ['int', 'float', 'str', 'numeric', 'list', 'dict'] and key not in filter_list:
            params.append('"{}": {}'.format(key, '<{}>'.format(key_type)))

    params.append('')

    if isinstance(func, PipelineSeed):
        return """results, summary = dsk.pipeline.auto({{'seed': '{}', 'params':{{{}}}}})""".format(func.name, param_tabs.join(params))

    if func.type == 'Feature Generator':
        return """dsk.pipeline.add_feature_generator([{{'name':'{}', 'params':{{{}}}}}])""".format(func.name, param_tabs.join(params))

    if func.type == 'Feature Selector':
        return """dsk.pipeline.add_feature_selector([{{'name':'{}', 'params':{{{}}}}}], params={{'number_of_features':(int)}})""".format(
            func.name, param_tabs.join(params))

    if func.type in ['Segmenter', 'Transform', 'Sampler']:
        return """dsk.pipeline.add_transform("{}", params={{{}}})""".format(func.name, param_tabs.join(params))

    if func.type == 'Training Algorithm':
        return """dsk.pipeline.set_training_algorithm("{}", params={{{}}})""".format(func.name, param_tabs.join(params))

    if func.type == 'Validation Method':
        return """dsk.pipeline.set_validation_method("{}", params={{{}}})""".format(func.name, param_tabs.join(params))

    if func.type == 'Classifier':
        return """dsk.pipeline.set_classifier("{}", params={{{}}})""".format(func.name, param_tabs.join(params))

    return None


class hold_function():
    def __init__(self, func):
        self.func = func

    def create_function(self):
        try:
            docs = function_help(self.func)
        except:
            docs = "unable to generate this function."
        try:
            get_ipython().set_next_input(docs, replace=True)
        except:
            print(docs)

        return docs


class Snippets():
    def __init__(self, function_df, function_list, seeds):
        self._function_df = function_df
        self._function_list = function_list
        self._seeds = seeds
        self._attach_types()

    def _attach_types(self):
        type_list = self._function_df.TYPE.unique()
        for func_type in type_list:
            setattr(self, func_type.replace(' ', '_'),
                    self._attach_functions(func_type))
        setattr(self, 'Auto', self._attach_seeds())

    def _attach_functions(self, func_type):
        funcs = self._function_df[self._function_df.TYPE == func_type]
        cls = empty()
        funcs = self._function_df[self._function_df.TYPE == func_type]
        for func in funcs.NAME:
            # functions names starting with numbers not allowed, check for them here
            sfunc = func
            if sfunc[:1] in '0123456789':
                sfunc = 'FG_' + func
            setattr(cls, sfunc.replace(' ', '_').replace('-', ''),
                    hold_function(self._function_list[func]).create_function)

        return cls

    def _attach_seeds(self):
        cls = empty()
        for seed in self._seeds:
            setattr(cls, seed.name.replace(' ', '_').replace('-', ''),
                    hold_function(seed).create_function)

        return cls


def get_group_columns(steps):
    for step in steps:
        inputs = step.get('inputs', None)
        if inputs:
            return inputs.get('group_columns')


def get_label_column(steps):
    for step in steps:
        if step.get('type', None) == 'tvo':
            return step.get('label_column', None)


def pipeline_function_help(function_list, step, filter_list=['group_columns', 'label_column', 'ignore_columns', 'passthrough_columns', "classifiers", 'validation_methods',  'input_data', 'return_segment_index']):

    function_name = step.get('name', None)
    if function_name is None:
        function_name = step.get('function_name', None)
    if function_name is None:
        return None

    func = function_list[function_name]

    params = []
    for item in func.input_contract:
        key = item.get('name')
        if key not in filter_list:
            value = step['inputs'].get(key, None)
            if value is not None:
                if item.get('type', None) == 'str':
                    params.append('"{}":"{}"'.format(key, value))
                else:
                    params.append('"{}":{}'.format(key, value))

    if func.type == 'Feature Generator':
        return """{{'name':'{}', 'params':{{{}}}}}""".format(func.name, param_tabs.join(params))

    if func.type == 'Feature Selector':
        return """{{'name':'{}', 'params':{{{}}}}}, """.format(func.name, param_tabs.join(params))

    if func.type in ['Segmenter', 'Transform', 'Sampler']:
        return """dsk.pipeline.add_transform("{}", params={{{}}})""".format(func.name, param_tabs.join(params))

    if func.type == 'Training Algorithm':
        return """dsk.pipeline.set_training_algorithm("{}", params={{{}}})""".format(func.name, param_tabs.join(params))

    if func.type == 'Validation Method':
        return """dsk.pipeline.set_validation_method("{}", params={{{}}})""".format(func.name, param_tabs.join(params))

    if func.type == 'Classifier':
        return """dsk.pipeline.set_classifier("{}", params={{{}}})""".format(func.name, param_tabs.join(params))

    return None


def build_pipeline(function_list, steps):
    pipeline_steps = ["dsk.pipeline.reset()"]

    for step in steps:
        if step.get('type') == 'generatorset':
            feature_generators = []
            for generator in step['set']:
                feature_generators.append(
                    pipeline_function_help(function_list, generator))
            pipeline_steps.append("""dsk.pipeline.add_feature_generator([{}])""".format(
                ',\n{}'.format(' '*40).join(feature_generators)))

        elif step.get('type') == 'selectorset':
            feature_selectors = []
            for selector in step['set']:
                feature_selectors.append(
                    pipeline_function_help(function_list, selector))
            pipeline_steps.append("dsk.pipeline.add_feature_selector([{}])""".format(
                ',\n{}'.format(' '*40).join(feature_selectors)))

        elif step.get('type') == 'tvo':
            for key in ['classifiers', 'validation_methods', 'optimizers']:
                pipeline_steps.append(pipeline_function_help(
                    function_list, step[key][0]))

            pipeline_steps.append("dsk.pipeline.set_tvo({{'validation_seed':{}}})".format(
                step.get('validation_seed', 0)))

        elif step.get('type', None) == 'featurefile':
            pipeline_steps.append("""dsk.pipeline.set_input_data("{}", data_columns={}, \n{}group_columns={},\n{}label_column="{}")""".
                                  format(step.get('name', None), step.get('data_columns', None), ' '*40, get_group_columns(steps), ' '*40, get_label_column(steps)))

        elif step.get('type', None) == 'query':
            pipeline_steps.append("""dsk.pipeline.set_input_query("{}")""".
                                  format(step.get('name', None)))
        else:
            pipeline_steps.append(pipeline_function_help(function_list, step))

    return pipeline_steps


def generate_pipeline(function_list, steps, replace=True):
    pipeline_steps = build_pipeline(function_list, steps)

    if replace:
        pipeline_string = '\n\n'.join([x for x in pipeline_steps if x])
        get_ipython().set_next_input(pipeline_string, replace=True)

    return pipeline_steps



from sensiml.render import BaseRenderer


class WidgetAttributeRenderer(BaseRenderer):

    def __init__(self, widget, render_attribute):
        self.widget = widget
        self.render_attribute = render_attribute

    def render(self, value):
        msg = value
        msg = '<b>' + msg + "</b>"
        msg = '<font size = "+0" >' + msg + "</font>"
        msg = '<p align="right";>' + msg + "</p>"
        setattr(self.widget, self.render_attribute, msg)



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


import json
from numpy import float64, int64
import sensiml.base.utility as utility
import logging
import time


logger = logging.getLogger(__name__)


class LabelType(object):
    Int = 'integer'
    Float = 'float'
    String = 'string'


class Label(object):
    """Base class for a label object."""

    def __init__(self, connection, project, capture):
        """Initialize a metadata object.

            Args:
                connection
                project
                capture
        """
        self._uuid = ''
        self._name = ''
        self._sample_start = 0
        self._sample_end = 0
        self._value = ''
        self._value_type = 'string'
        self._comments = ''
        self._connection = connection
        self._project = project
        self._capture = capture
        self._lastmodifiedby = ''
        self._lastmodifiedtime = ''
        self._segmenter = None

    @property
    def uuid(self):
        """Auto generated unique identifier for the metadata object"""
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def name(self):
        """The name property of the metadata object"""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def value(self):
        """The value of the metadata object

        Note:
            The metadata value -1 (as an integer or string) has a special meaning. It is reserved by KnowledgeBuilder
            and means that the metadata name is not defined for a particular sample. You should never use -1 for a
            metadata value.
        """
        if self._value_type == LabelType.Float:
            return float(self._value)
        elif self._value_type == LabelType.Int:
            return int(self._value)
        else:
            return self._value

    @value.setter
    def value(self, value):
        if type(value) in (float, float64):
            self._value_type = LabelType.Float
            self._value = str(value)
        elif type(value) in (int, int64):
            self._value_type = LabelType.Int
            self._value = str(value)
        else:
            self._value_type = LabelType.String
            self._value = str(value)

    @property
    def value_type(self):
        """The data type of the metadata object"""
        return self._value_type

    @property
    def sample_start(self):
        """The index of the first sample of the label"""
        return self._sample_start

    @sample_start.setter
    def sample_start(self, value):
        self._sample_start = value

    @property
    def sample_end(self):
        """The index of the last sample of the label"""
        return self._sample_end

    @sample_end.setter
    def sample_end(self, value):
        self._sample_end = value

    @property
    def segmenter(self):
        """The index of the last sample of the label"""
        return self._segmenter

    @segmenter.setter
    def segmenter(self, value):
        self._segmenter = value

    @property
    def lastmodifiedby(self):
        return self._lastmodifiedby

    @property
    def lastmodifiedtime(self):
        return self._lastmodifiedtime

    @property
    def comments(self):
        """The label comments"""
        return self._comments

    @comments.setter
    def comments(self, value):
        self._comments = value

    def insert(self):
        """Calls the REST API and inserts a metadata object onto the server using the local object's properties."""
        self._capture.await_ready()
        url = 'project/{0}/capture/{1}/label-relationship/'.format(
            self._project.uuid, self._capture.uuid)
        label_info = {'name': self.name,
                      'type': self.value_type,
                      'capture_sample_sequence_start': self.sample_start,
                      'capture_sample_sequence_end': self.sample_end,
                      'value': self.value,
                      'segmenter': self.segmenter,
                      'comments': self._comments}
        response = self._connection.request('post', url, label_info)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.uuid = response_data['uuid']

    def update(self):
        """Calls the REST API and updates the object on the server."""
        self._capture.await_ready()
        url = 'project/{0}/capture/{1}/label-relationship/{2}/'.format(
            self._project.uuid, self._capture.uuid, self.uuid)
        label_info = {'name': self.name,
                      'type': self.value_type,
                      'capture_sample_sequence_start': self.sample_start,
                      'capture_sample_sequence_end': self.sample_end,
                      'value': self.value,
                      'segmenter': self.segmenter,
                      'comments': self._comments}
        response = self._connection.request('put', url, label_info)
        response_data, err = utility.check_server_response(response)

    def delete(self):
        """Calls the REST API and deletes the object from the server."""
        url = 'project/{0}/capture/{1}/label-relationship/{2}/'.format(
            self._project.uuid, self._capture.uuid, self.uuid)
        response = self._connection.request('delete', url,)
        response_data, err = utility.check_server_response(response)

    def refresh(self):
        """Calls the REST API and populates the local object's properties from the server."""
        url = 'project/{0}/capture/{1}/label-relationship/{2}/'.format(
            self._project.uuid, self._capture.uuid, self.uuid)
        response = self._connection.request('get', url,)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.initialize_from_dict(response_data)

    def initialize_from_dict(self, dict):
        """Reads a json dictionary and populates a single metadata object.

            Args:
                dict (dict): contains the uuid, name, type, value, capture_sample_sequence_start, and
                capture_sample_sequence_end properties
        """
        self.uuid = dict['uuid']
        self.name = dict['name']
        self.sample_start = dict['capture_sample_sequence_start']
        self.sample_end = dict['capture_sample_sequence_end']
        self.segmenter = dict.get('segmenter', None)
        if dict['type'] == LabelType.Int:
            self.value = int(dict['value'])
        elif dict['type'] == LabelType.Float:
            self.value = float(dict['value'])
        else:
            self.value = dict['value']



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


import json
from  pandas import DataFrame
import re
import warnings
import types

import sensiml.base.utility as utility


class QueryColumns(object):
    """Base class for the query columns and metadata_columns properties"""
    def __init__(self, columns=[]):
        if len(columns) > 0:
            self._columns = columns
        else:
            self._columns = []

    def add(self, *items):
        """Adds a column name or names for query selection.

            Args:
                items (str or list[str]): column names to add

            Note:
                The named column(s) must exist in the project and duplicate column names will be ignored
        """
        for item in items:
            if item not in self._columns:
                self._columns.append(item)

    def remove(self, item):
        """Removes a column name from query selection.

            Args:
                item (str): column name to remove
        """
        if item in self._columns:
            self._columns.remove(item)

    def clear(self):
        """Clears all column names from query selection"""
        self._columns = []

    def __str__(self):
        return json.dumps(self._columns)

class Query(object):
    """Base class for a query.

    Queries extract project data, or a subset of project data, for use in a pipeline. The query must specify which
    columns of data to extract and what filter conditions to apply, and KnowledgeBuilder takes care of the database
    joins required to produce a clean, analysis-ready data frame.
    """
    def __init__(self, connection, project):
        """Initializes a query instance"""
        self._uuid = ''
        self._name = ''
        self._columns = QueryColumns()
        self._metadata_columns = QueryColumns()
        self._metadata_filter = ''
        self._segmenter = None
        self._label_column = ''
        self._combine_labels = None
        self._dirty = False
        self._connection = connection
        self._project = project

    @property
    def uuid(self):
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def name(self):
        """Name of the query"""
        return self._name

    @name.setter
    def name(self, value):
        self._dirty = True
        self._name = value



    @property
    def columns(self):
        """Sensor columns to include in the query result

        Note:
            Columns must correspond to actual project sensor columns or the reserved word 'SequenceID' for the
            original sample index.
        """
        self._dirty = True
        return self._columns

    @property
    def label_column(self):
        """Label columns to use in the query result

        Note:
            Columns must correspond to actual project label column"""

        self._dirty = True
        return self._label_column


    @label_column.setter
    def label_column(self, value):
        self._dirty = True
        self._label_column = value

    @property
    def combine_labels(self):
        """Combine label values into new value to use in the query result

        Label = Gesture
        Label_Values = A,B,C,D,E
        combine_labels = {'Group1':['A','B',C'],'Group2':['D','E']}

        the labels that will be returned will be group1 and group2
        """

        self._dirty = True
        return self._combine_labels


    @combine_labels.setter
    def combine_labels(self, value):
        self._dirty = True
        self._combine_labels = value

    @property
    def metadata_columns(self):
        """Metadata columns to include in the query result

        Note:
            Columns must correspond to actual project metadata columns.
        """
        self._dirty = True
        return self._metadata_columns

    @property
    def metadata_filter(self):
        """Filter criteria of the query

        Args:
            value (str): similar to a SQL WHERE clause, the string can contain any number of AND-concatenated
              expressions where square brackets surround the column name and comparison value, with the operator in
              between. Supported operators: >, >=, <, <=, =, !=, IN

        Examples::

            metadata_filter = '[Subject] > [5] AND [Subject] <= [15]'
            metadata_filter = '[Gender] = [Female] AND [Activity] != [Walking]'
            metadata_filter = '[Subject] IN [5, 7, 9, 11, 13, 15]'

        Note:
            Queries do not support OR-concatenation between expressions, but often the IN operator can be used to
            achieve OR-like functionality on a single column. For example::

                [Gesture] IN [A, M, L]

            is equivalent to::

                [Gesture] = [A] OR [Gesture] = [M] OR [Gesture] = [L]

        """
        return self._metadata_filter

    @metadata_filter.setter
    def metadata_filter(self, value):
        self._dirty = True
        self._metadata_filter = value

    @property
    def segmenter(self):
        """Segmenter to use for the query

        Args:
            value (int): ID of segmenter.

        """
        return self._segmenter

    @segmenter.setter
    def segmenter(self, value):
        if value is None or isinstance(value, int):
            self._segmenter = value
        else:
            raise ValueError('Value must be a integer')

    def insert(self):
        """Calls the REST API and inserts a new query."""
        url = 'project/{0}/query/'.format(self._project.uuid)

        query_info = {'name': self.name, 'columns': self.columns._columns,
                      'metadata_columns': self.metadata_columns._columns,
                      'metadata_filter': self.metadata_filter,
                      'segmenter_id': self._segmenter,
                      'label_column': self.label_column,
                      'combine_labels': self.combine_labels}

        request = self._connection.request('post', url, query_info)
        response, err = utility.check_server_response(request)
        if err is False:
            self.uuid = response['uuid']
            self._dirty = False


    def update(self):
        """Calls the REST API and updates the query object on the server."""
        url = 'project/{0}/query/{1}/'.format(self._project.uuid, self.uuid)

        self._checked = None
        query_info = {'name': self.name, 'columns': self.columns._columns,
                        'metadata_columns': self.metadata_columns._columns,
                        'metadata_filter': self.metadata_filter,
                        'segmenter_id': self._segmenter,
                        'label_column': self.label_column,
                        'combine_labels': self.combine_labels}

        if (len(query_info['columns']) > 0 or len(query_info['metadata_columns']) > 0):
            filters = self._metadata_filter.split('AND') if len(self._metadata_filter) > 0 else []
            for i, filter in enumerate([filter for filter in filters if len(filters)]):
                match = re.search("\[(?P<key>[0-9A-Za-z_\-\. \)\(]+)\]([ ]*(?P<symbol>[\>\=|\<\=|\=|\<\>|\>|\<|in|IN|!=]+)[ ]*\[(?P<value>.+)\])?", str(filter))
                if not match:
                    self._checked = 1
            if self._checked is None:
                request = self._connection.request('put', url, query_info)
                response, err = utility.check_server_response(request)
            else:
                self._metadata_filter = ''
                print("Metadata Filter is not formatted correctly!")
        else:
            print( "Sensor columns and, or metadata columns must be specified in order to query!")

    def delete(self):
        """Calls the REST API and deletes the query object from the server."""
        url = 'project/{0}/query/{1}/'.format(self._project.uuid, self.uuid)
        request = self._connection.request('delete', url,)
        response, err = utility.check_server_response(request)
        self._dirty = False
        if err is False:
            print(response['result'])

    def statistics_segments(self):
        """Returns metadata statistics for the query."""
        if self._dirty:
            self.update()

        url = 'project/{0}/query/{1}/statistics/'.format(self._project.uuid, self.uuid)
        request = self._connection.request('get', url,)
        response_data, err = utility.check_server_response(request)
        if err is False:
            return DataFrame(response_data)
        else:
            return None

    def statistics(self):
        """Returns metadata statistics for the query."""
        if self._dirty:
            self.update()

        url = 'project/{0}/query/{1}/statistics/'.format(self._project.uuid, self.uuid)
        request = self._connection.request('get', url,)
        response_data, err = utility.check_server_response(request)
        if err is False:
            df = DataFrame(response_data)
            df.drop('Segment Length', axis=1, inplace=True)
            data = df.groupby(['Capture']+self.metadata_columns._columns+[self.label_column])[self.label_column].size().unstack(fill_value=0)

            return data
        else:
            return None



    def data(self):
        """Calls the REST API for query execution and returns the result.

        Note:
            Intended for previewing the query result before creating a query call object and using it in a sandbox
            step. The resulting DataFrame is not cached on the server, but once it is used in a sandbox it may be
            cached.
        """
        warnings.warn('This call has a timeout of two minutes.')

        if self._dirty:
            self.update()

        url = 'project/{0}/query/{1}/data/'.format(self._project.uuid, self.uuid)
        request = self._connection.request('get', url,)
        response_data, err = utility.check_server_response(request)
        if err is False:
            error_report = None
            if not isinstance(response_data, list):
                data = DataFrame(response_data)
            else:
                try:
                    data = DataFrame(response_data[0])
                    error_report = response_data[1]
                except:
                    data = DataFrame(response_data)

            # Post-processing to get numeric index
            data['sample'] = data.index.astype(int)
            data = data.set_index('sample')
            data.index.names = [None]
            data = data.sort_index(axis=0)
            cols = data.columns.tolist()
            if 'id' in cols:
                data = data[['id'] + [c for c in cols if c != 'id']]

            return data.sort_index(axis=0), error_report
        else:
            return None

    def size(self):
        """Returns the size of the dataframe which would result from the query."""
        # Sync the client-side query with server-side query object before we check the size
        if self._dirty:
            self.update()

        url = 'project/{0}/query/{1}/size/'.format(self._project.uuid, self.uuid)
        request = self._connection.request('get', url,)
        response_data, err = utility.check_server_response(request)
        if not err:
            query_result = response_data
        else:
            query_result = None
        return query_result

    def refresh(self):
        """Calls the REST API and self populate using the uuid."""
        url = 'project/{0}/query/{1}/'.format(self._project.uuid, self.uuid)
        request = self._connection.request('get', url,)
        response, err = utility.check_server_response(request)
        if err is False:
            self.uuid = response['uuid']
            self.name = response['name']
            self._columns = QueryColumns(response['columns'])
            self._label_column = response['label_column']
            self._metadata_columns = QueryColumns(response['metadata_columns'])
            self.metadata_filter = response['metadata_filter']
            self.segmenter = response['segmenter']
            self.combine_labels = response['combine_labels']
            self._dirty = False

    def initialize_from_dict(self, query):
        """Reads a json dict and populates a single query."""
        self.uuid = query['uuid']
        self.name = query['name']
        self._columns = QueryColumns(query['columns'])
        self._metadata_columns = QueryColumns(query['metadata_columns'])
        self.metadata_filter = query['metadata_filter']
        self.label_column = query.get('label_column', '')
        self.segmenter = query.get('segmenter', None)
        self.combine_labels = query.get('combine_labels', None)
        self._dirty = False


    def __str__(self):
        return ('NAME: {0}\n'
                'COLUMNS: {1}\n'
                'METADATA_COLUMNS: {2}\n'
                'METADATA FILTER: {3}\n'
                'SEGMENTER ID: {4}\n'
                'LABEL COLUMN {5}\n'
                'COMBINE LABELS {6}\n'
                ).format(self.name, self.columns,
                         self.metadata_columns, self.metadata_filter,
                         self.segmenter, self.label_column, self.combine_labels)

    def _to_dict(self):
        return {'name': self.name,
            'sensor_columns': self.columns._columns,
            'metadata_columns': self.metadata_columns._columns,
            'metadata_filter': self.metadata_filter,
                'segmenter_id': self.segmenter,
                'label_column': self.label_column}



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017-18  SensiML Corporation.                                   #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import NullFormatter
import matplotlib.patches as patches
from copy import deepcopy
from seaborn import boxplot, violinplot, set_style
from pandas import DataFrame, concat
import io


class Visualize(object):

    def __init__(self, feature_vector=None, model=None, label=None):
        """
                model: model result or knowledpack
                feature_vector (Dataframe): feature set that is results of generator_set or feature_selector
                label (String), name of class
        """
        self._feature_vector = None
        self._knowledgepack = None
        self._label = None

        self.knowledgepack = model
        self.label = label
        if self.label is None:
            print(
                "Label column is not specified for this knowledpack, specify as label=XXX")
        self.feature_vector = feature_vector

    @property
    def knowledgepack(self):
        """Auto generated unique identifier for the project object"""
        return self._knowledgepack

    @knowledgepack.setter
    def knowledgepack(self, value):
        if value is None:
            self._knowledgepack = None
        elif hasattr(value, 'knowledgepack'):
            self._knowledgepack = value.knowledgepack
        else:
            self._knowledgepack = value

    @property
    def label(self):
        """Auto generated unique identifier for the project object"""
        return self._label

    @label.setter
    def label(self, value):
        if value is None and self.knowledgepack:
            self._label = self.knowledgepack.query_summary.get(
                'label_column', None)
        else:
            self._label = value

    @property
    def feature_vector(self):
        """Auto generated unique identifier for the project object"""
        return self._feature_vector

    @feature_vector.setter
    def feature_vector(self, value):
        self._feature_vector = value[[
            i for i in sorted(value.columns.tolist()) if ('gen_' == i[:4])]+[self.label]]

    def plot_features(self, features_list=None,
                      plot_style='boxplot', figsize=(20, 5),
                      features_per_row=4):
        """
        Display the distributions of feature values across classes.

        Args:
            features (Dataframe): Feature set that is results of generator_set or feature_selector
            label (String): Name of class label
            features_list (List):  List of features to plot, if None plots all (default: None)
            plot_style (String):  'violinplot' / 'boxplot' to determine style of plot
            figsize (tuple): Size of figure to make
            features_per_row (int): Number of features to show in each row of the plot
        """

        # if features_list is not given use all featues
        if not(features_list):
            features_list = [
                i for i in self.feature_vector.columns.tolist() if ('gen_' == i[:4])]

        # reframe the feature dataframe so it can be used to input for violinplot or boxplot
        num_plots = len(features_list)//(features_per_row)
        if len(features_list) % features_per_row != 0:
            num_plots += 1
        figsize = (figsize[0], figsize[1]*num_plots)
        f, axarr = plt.subplots(num_plots, figsize=figsize)
        set_style('ticks')
        if not isinstance(axarr, np.ndarray):
            axarr = [axarr]

        for index in range(num_plots):
            df_list = [reframe(self.feature_vector, feature, self.label) for feature in
                       features_list[index*features_per_row:(index+1)*features_per_row]]

            if plot_style == 'violinplot':
                violinplot(x="Features", y="Sensor Values", hue=self.label,
                           data=concat(df_list), palette="muted", ax=axarr[index])
            elif plot_style == 'boxplot':
                boxplot(x="Features", y="Sensor Values", hue=self.label,
                        data=concat(df_list), palette="muted", ax=axarr[index])
            else:
                print('Error: plot_style is not in the library')
                return

            axarr[index].set_xticklabels(
                features_list[index *
                              features_per_row:(index+1)*features_per_row],
                fontsize=16)
            axarr[index].set_ylabel('Feature Values', fontsize=16)
            axarr[index].set_xlabel('')
            axarr[index].legend()
            axarr[index].set_ylim((0, 255))
            axarr[index].legend().set_visible(False)

        axarr[-1].set_xlabel('Features', fontsize=16)
        axarr[0].legend(loc='center left',
                        bbox_to_anchor=(1, 0.85), fontsize=16)
        axarr[0].set_title("Feature Vector Distribution", fontsize=16)

        plt.show()

    def neuron_feature_map(self,
                           feature1,
                           features_list,
                           label=None,
                           norm=None,
                           hist_figure=True,
                           neuron_figure=True,
                           vector_labels=['all'],
                           vector_alpha=0.3,
                           neuron_alpha=0.2,
                           color_list=['m', 'k', 'b', 'lime', 'r', 'c', 'y', 'brown', 'tan', 'g', 'pink', 'gray']):
        """
            This function display neuron map, feature map and feature distributions.

            Args:
                hist_figure = Bool, display historgram of vector for given features.
                neuron_figure = True,
                pivot_feature = String, it is y-axis feature that used to display neuron map, feature map and feature distributions
                versus_features = List<String>, x-axis features that used to display neuron map, feature map and feature distributions
                vector_labels (List): which features will be displayed  [ f1, f2, f3, ..., 'all', [] ], ['all'] display all features, [] does not display any features
                vector_alpha = 0.3, level of vector transparency
                neuron_alpha = 0.1, level of neuron transparency
                norm(int): takes from knowledpack, to overide 0=L1, 1=LSUP

            Returns: None

        """

        self.hist_figure = hist_figure
        self.pivot_feature = feature1
        self.vector_labels = vector_labels
        self.vector_alpha = vector_alpha
        self.neuron_alpha = neuron_alpha
        self.axis = 10  # axis
        self.color_list = color_list

        if isinstance(features_list, str):
            features_list = [features_list]

        self.columns_index = [None, None]
        self.columns_index[1] = self.feature_vector.columns.tolist().index(
            self.pivot_feature)

        for feature in features_list:
            self.columns_index[0] = self.feature_vector.columns.tolist().index(
                feature)

            if self.vector_labels == ['all']:
                self.vector_labels = np.unique(
                    self.feature_vector[self.label]).tolist()

            self._set_initial_figure()

            if neuron_figure and self.knowledgepack:
                if norm is None:
                    norm = self.knowledgepack.device_configuration.get(
                        'distance_mode', 0)
                else:
                    norm = norm
                self._plot_neurons(
                    self.knowledgepack.neuron_array, feature, norm)

            if (vector_labels != ['None']) | (vector_labels == []):
                self._plot_feature_vector(feature)

            self._set_final_figure_settings(feature)
            plt.show()

    def _plot_neurons(self, neurons, versus_feature, norm):
        num = 0
        for neuron in neurons:
            x_val = neuron['Vector'][self.columns_index[0]]
            y_val = neuron['Vector'][self.columns_index[1]]
            aif = neuron['AIF']

            if neuron['Category'] < 30000:  # if it is not degenerated neuron
                cat_label = self.knowledgepack.class_map[str(
                    neuron['Category'])]
                clr = self.color_list[self.label_indx_list.index(cat_label)]
                if norm == 0:
                    dim = len(neuron['Vector'])
                    new_aif = aif * np.sqrt(2)/dim
                    self.neuron_scatter.add_patch(patches.Rectangle((x_val-(aif*2/dim), y_val),
                                                                    2*new_aif,
                                                                    2*new_aif,
                                                                    color=clr,
                                                                    alpha=self.neuron_alpha,
                                                                    angle=-45.0))

                else:
                    self.neuron_scatter.add_patch(patches.Rectangle((x_val-aif, y_val-aif),
                                                                    2*aif,
                                                                    2*aif,
                                                                    color=clr,
                                                                    alpha=self.neuron_alpha,
                                                                    angle=0))

                self.neuron_scatter.scatter(
                    x_val, y_val, c=clr, marker='*', s=96)

    def _plot_hist(self, Xs, Ys, clr):
        binwidth = 5
        bins = np.arange(0, 256, binwidth)
        self.axHistx.hist(Xs, bins=bins, color=clr, alpha=self.vector_alpha)
        self.axHistx.set_xlim(self.neuron_scatter.get_xlim())

        self.axHisty.hist(Ys, bins=bins, color=clr,
                          alpha=self.vector_alpha, orientation='horizontal')
        self.axHisty.set_ylim(self.neuron_scatter.get_ylim())

    def _plot_feature_vector(self, versus_feature):
        for label_indx in self.vector_labels:
            temp_label_data = self.feature_vector[self.feature_vector[self.label] == label_indx].index.tolist(
            )
            temp = self.feature_vector.loc[temp_label_data[0]].tolist()
            cat_label = temp[-1]
            clr = self.color_list[self.label_indx_list.index(cat_label)]
            Xs = self.feature_vector.loc[temp_label_data,
                                         versus_feature].tolist()
            Ys = self.feature_vector.loc[temp_label_data,
                                         self.pivot_feature].tolist()

            self.neuron_scatter.scatter(Xs, Ys, c=clr, alpha=self.vector_alpha)

            if self.hist_figure:
                self._plot_hist(Xs, Ys, clr)

    def _set_initial_figure(self):
        plt.figure(1, figsize=(self.axis, self.axis))

        left, width = 0.1, 0.65
        bottom, height = 0.1, 0.65
        rect_scatter = [left, bottom, width, height]
        self.neuron_scatter = plt.axes(rect_scatter, aspect='equal')

        if self.hist_figure:
            # definitions for the axes
            bottom_h = left_h = left + width + 0.02
            rect_histx = [left, bottom_h, width, 0.2]
            rect_histy = [left_h, bottom, 0.2, height]
            nullfmt = NullFormatter()
            self.axHistx = plt.axes(rect_histx)
            self.axHistx.xaxis.set_major_formatter(nullfmt)
            self.axHisty = plt.axes(rect_histy)
            self.axHisty.yaxis.set_major_formatter(nullfmt)

        if self.knowledgepack:
            self.label_indx_list = np.unique([self.knowledgepack.class_map[str(i)] for i in
                                              np.sort(list(map(int, list(self.knowledgepack.class_map.keys()))))] +
                                             np.unique(self.feature_vector[self.label]).tolist()).tolist()
        else:
            self.label_indx_list = list(
                np.unique(self.feature_vector[self.label]))

        # this part creates dummy vectors with the unique number of labels for legent.
        for indx in range(len(self.label_indx_list)):
            self.neuron_scatter.scatter(
                0, 0, c=self.color_list[indx], label=self.label_indx_list[indx])

        self.neuron_scatter.set_xlim([0, 256])
        self.neuron_scatter.set_ylim([0, 256])

    def _set_final_figure_settings(self, versus_feature):

        self.neuron_scatter.set_xlabel(versus_feature, fontsize=12)
        self.neuron_scatter.set_ylabel(self.pivot_feature, fontsize=12)

        self.neuron_scatter.set_xticks(range(0, 256, 10))
        self.neuron_scatter.set_xticklabels(
            range(0, 256, 10), rotation=0, fontsize=8)

        self.neuron_scatter.set_yticks(range(0, 256, 10))
        self.neuron_scatter.set_yticklabels(
            range(0, 256, 10), rotation=0, fontsize=8)

        if self.hist_figure:
            self.axHistx.set_xticks(range(0, 256, 10))
            self.axHistx.yaxis.set_tick_params(labelsize=8)
            self.axHistx.set_title(
                self.pivot_feature + ' vs ' + versus_feature)

            self.axHisty.set_yticks(range(0, 256, 10))
            self.axHisty.xaxis.set_tick_params(labelsize=8)
            self.neuron_scatter.legend(
                loc='center left', bbox_to_anchor=(1.02, 1.26), fontsize=12)

        else:
            self.neuron_scatter.legend(
                loc='center left', bbox_to_anchor=(1, 0.945), fontsize=12)
            self.neuron_scatter.set_title(
                self.pivot_feature + ' vs ' + versus_feature)


def reframe(features, feature, label):
    # This funciton reframes the original features dataframe
    temp_list = DataFrame(columns=[label, 'Features', 'Sensor Values'])
    temp_list[label] = features[label].tolist()
    temp_list['Features'] = [feature] * len(features)
    temp_list['Sensor Values'] = features[feature].tolist()
    return temp_list



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


from sensiml.method_calls.generatorcall import GeneratorCall
from sensiml.datamanager.pipeline import PipelineStep


class GeneratorCallSet(PipelineStep):
    """The base class for a collection of generator calls"""

    def __init__(self, name=''):
        super(GeneratorCallSet, self).__init__(
            name=name, step_type='GeneratorCallSet')
        self._input_data = ''
        self._generator_calls = []
        self._group_columns = None

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def input_data(self):
        return self._input_data

    @input_data.setter
    def input_data(self, value):
        self._input_data = value

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, value):
        self._outputs = value

    @property
    def generator_calls(self):
        return self._generator_calls

    def add_generator_call(self, *generator_call):
        """Adds one or more generator calls to the collection.

            Args:
                generator_call (GeneratorCall or list[GeneratorCall]): object(s) to append
        """
        for g in generator_call:
            self._generator_calls.append(g)

    def remove_generator_call(self, *generator_calls):
        """Removes a generator call from the collection.

            Args:
                generator_calls (GeneratorCall): object to remove
        """
        for generator_call in generator_calls:
            self._generator_calls = [
                f for f in self._generator_calls if f != generator_call]
        # Note:  We should probably re-populate inputs and outputs here

    @property
    def group_columns(self):
        return self._group_columns

    @group_columns.setter
    def group_columns(self, value):
        self._group_columns = value

    def _to_list(self):
        gencalls = []
        for item in self._generator_calls:
            gencalls.append(item._to_dict())
        return gencalls

    def _to_dict(self):
        gencalls_set = []
        set_dict = {}
        for item in self._generator_calls:
            gencalls_set.append(item._to_dict())
        set_dict['name'] = self._name
        set_dict['type'] = 'generatorset'
        set_dict['set'] = gencalls_set
        set_dict['inputs'] = {}
        set_dict['inputs']['group_columns'] = self._group_columns
        set_dict['inputs']['input_data'] = self._input_data
        set_dict['outputs'] = self._outputs
        return set_dict



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################

import os
import json
import sensiml.base.utility as utility


class FeatureFile(object):
    """Base class for a featurefile object."""

    def __init__(self, connection, project, name='', path=''):
        self._connection = connection
        self._project = project
        self.uuid = None
        self.name = name
        self.path = path

    # Maintain campatibility with old filename attr
    @property
    def filename(self):
        """The name of the file as stored on the server

        Note:
            Filename must contain a .csv or .arff extension
        """
        return self.name

    @filename.setter
    def filename(self, value):
        self.name = value

    def insert(self):
        """Calls the REST API to insert a new featurefile."""
        url = 'project/{0}/featurefile/'.format(
            self._project.uuid)
        featurefile_info = {'name': self.name}

        if not os.path.exists(self.path):
            raise OSError("Cannot update featurefile because the system cannot find the file path: '{0}'. "
                          "Please update the file's path attribute.".format(self.path))

        response = self._connection.file_request(url,
                                                 self.path, featurefile_info, "rb")
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.uuid = response_data['uuid']

    def update(self):
        """Calls the REST API to update the featurefile's properties on the server."""
        url = 'project/{0}/featurefile/{1}/'.format(
            self._project.uuid, self.uuid)
        featurefile_info = {'name': self.name}

        if not os.path.exists(self.path):
            raise OSError("Cannot update featurefile because the system cannot find the file path: '{0}'. "
                          "Please update the file's path attribute.".format(self.path))

        response = self._connection.file_request(url,
                                                 self.path, featurefile_info, "rb", method='put')
        response_data, err = utility.check_server_response(response)

    def delete(self):
        """Calls the REST API and deletes the featurefile from the server."""
        url = 'project/{0}/featurefile/{1}/'.format(
            self._project.uuid, self.uuid)
        response = self._connection.request('delete', url,)
        response_data, err = utility.check_server_response(response)

    def refresh(self):
        """Calls the REST API and populate the featurefile's properties from the server."""
        url = 'project/{0}/featurefile/{1}/'.format(
            self._project.uuid, self.uuid)
        response = self._connection.request('get', url,)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.name = response_data['name']

    def data(self):
        """Calls the REST API and retrieves the featurefile's binary data.

        Returns:
            featurefile contents
        """
        url = 'project/{0}/featurefile/{1}/data/'.format(
            self._project.uuid, self.uuid)
        response = self._connection.request(
            'get',
            url,
        )
        return response.content

    def initialize_from_dict(self, dict):
        """Makes a new featurefile using the dictionary.

        Args:
            dict (dict): contains the featurefile's 'name' and 'uuid' properties
        """
        self.uuid = dict['uuid']
        self.name = dict['name']



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################

import os
import json
import sensiml.base.utility as utility


class Segmenters(object):
    """Base class for a segmenter object."""

    def __init__(self, connection, project):
        self._connection = connection
        self._project = project

    def get_segmenters(self):
        """Gets a list of all segmenters in the project.

        Returns:
            list (segmenters)
        """
        err = False
        url = 'project/{0}/segmenter/'.format(self._project.uuid)
        response = self._connection.request('get', url,)
        try:
            response_data, err = utility.check_server_response(response)
        except ValueError:
            print(response)
        # Populate the retrieved featurefiles
        segmenters = []

        if err is False:
            try:
                for segmenter_params in response_data:
                    segmenters.append(segmenter_params)
            except:
                pass

        return segmenters



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017-18  SensiML Corporation.                                   #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


import sys
import os
import glob
import serial
from os import listdir, system, makedirs
from subprocess import Popen, PIPE
from os.path import isfile, join, abspath, dirname, exists, splitext
from ipywidgets import widgets
from ipywidgets import Layout, Button, VBox, HBox, Textarea, Dropdown, Text, Button
from IPython.display import display
from ipywidgets import IntText
from sensiml.widgets.base_widget import BaseWidget
import zipfile

if sys.platform == 'win32':
    nrfutil = os.path.join(os.path.dirname(sys.executable), 'nrfutil.exe')
    if not os.path.exists(nrfutil):
        nrfutil = os.path.join(os.path.dirname(
            sys.executable), 'Scripts', 'nrfutil.exe')
else:
    nrfutil = os.path.join(os.path.dirname(sys.executable), 'nrfutil')


# 11/21/17: Is there a better way to do this? Add it to the model as well?
extensions_by_platform = {
    'Nordic Thingy 52': '.zip'
}

commands_by_platform = {
    'Nordic Thingy 52': {
        # 'OTA': r'{} pkg generate --application "{}" --application-version 1 --hw-version 52 --sd-req 0x98 --key-file "{}" "{}"',
        'Flash': {
            'Jlink': 'nrfjprog --program "{}" -f nrf52 --sectorerase && nrfjprog --reset -f nrf52',
            # 'BLE-DFU' : 'nrfutil dfu ble -ic NRF52 -pkg {} -p {} -n "ThingyDfu" ',
            # 'Serial' : 'nrfutil dfu serial -pkg {} -p {}'
        }
    },

}


SUPPORTED_BOARDS = ["Nordic Thingy"]


class FlashWidget(BaseWidget):
    def __init__(self, dsk=None, folder='knowledgepacks'):
        self._dsk = dsk
        self._folder = folder
        self._flashing_file = ''
        self.platforms = None
        if not exists(folder):
            makedirs(folder)

    def _refresh_platform_files(self, b=None):
        if self._dsk is None:
            return

        key_name = self.platform_name.replace(' ', '-')
        platform_search = '{}_{}'.format(
            key_name, self.platform_version).lower()
        ext_match = extensions_by_platform.get(self.platform_name, '.zip')

        platform_files = [f for f in listdir(self._folder) if isfile(join(self._folder, f))
                          and platform_search in f.lower() and splitext(f)[-1] == ext_match]

        self.flashable_files_widget.options = platform_files
        if len(platform_files) > 0:
            self.flashable_files_widget.value = platform_files[0]

    def _refresh_flash_methods(self):
        flash_methods = []
        if commands_by_platform.get(self.platform_name, None):
            flash_methods = list(
                commands_by_platform[self.platform_name]['Flash'].keys())
        self.flash_methods_widget.options = flash_methods

    def select_platform(self, platform_name):
        if self.platforms is None:
            return

        self.kp_platform_id = self._dsk.platforms[self.platform_widget.value].id
        self.platform_name = self.platforms.iloc[platform_name]['Software Platform']
        self.platform_version = self.platforms.iloc[platform_name]['Platform Version']

        self._refresh_platform_files()
        self._refresh_flash_methods()

        # if self._dsk.platforms[self.platform_widget.value].ota_capable:
        #    self.ota_button.layout.visibility = "visible"
        # else:
        self.ota_button.layout.visibility = "hidden"

        self.description_widget.value = self.platforms.iloc[platform_name]['Description']

    def get_platform_names(self):
        pf = {}
        for i in range(len(self.platforms)):
            if self.platforms.iloc[i]['Board'] in SUPPORTED_BOARDS:
                pf['{} {}'.format(self.platforms.iloc[i]['Board'],
                                  self.platforms.iloc[i]['Platform Version'])] = i

        return pf

    def _wait_for_process(self, command):
        print('Running the following command: ')
        print(command)
        sub_proc = Popen(command[0], stdout=PIPE, stderr=PIPE, shell=True)
        p_val = sub_proc.poll()
        while(p_val is None):
            print(sub_proc.stdout.read())
            print(sub_proc.stderr.read())
            p_val = sub_proc.poll()

        if p_val is 0:
            print("Process Completed successfully.")
        else:
            print("And error occured. Error code {}".format(p_val))

    def on_flash_button_clicked(self, b):
        if self.platforms is None:
            return
        if not self.flashable_files_widget.value:
            return

        available_commands = commands_by_platform[self.platform_name]['Flash']

        if 'Nordic' in self.platform_name:
            self._flashing_file = unzip_folder(
                abspath(join(self._folder, self.flashable_files_widget.value)))
            command = [available_commands['Jlink'].format(self._flashing_file)]

        else:
            print('Platform is not yet supported.')
            return

        self._wait_for_process(command)

    def on_generate_button_clicked(self, b=None):
        if self.platforms is None:
            return
        if not self.flashable_files_widget.value:
            return

        print('Generating OTA Package for platform{}'.format(self.platform_name))

        flash_file = unzip_folder(
            abspath(join(self._folder, self.flashable_files_widget.value)))
        output_file = abspath(join(self._folder, '{}.zip'.format(flash_file)))
        self.kp_platform_id = self._dsk.platforms[self.platform_widget.value].id

        if 'Thingy' in self.platform_name:  # This is Nordic Thingy
            keyfile = join(os.path.dirname(flash_file),
                           'sensiml_example_key.pem')
            write_key_file(keyfile)
            command = [commands_by_platform['Nordic Thingy 52']['OTA'].format(
                nrfutil, flash_file, keyfile, output_file)]
            self._wait_for_process(command)
        else:
            print('Platform not supported for generating OTA package')

    def _refresh(self, b=None):
        if self._dsk is None:
            return
        self.platforms = self._dsk.platforms()
        self.platform_widget.options = self.get_platform_names()

    def create_widget(self):
        self.platform_widget = widgets.Dropdown(description='Platform')
        self.description_widget = widgets.Label(layout=Layout(width='66%'))
        self.flashable_files_widget = widgets.Dropdown(
            description='Binary', layout=Layout(width='85%'))
        self.flash_methods_widget = widgets.Dropdown(
            description="Flash Method")
        self.ota_button = Button(description="Generate OTA")
        self.flash_button = Button(description="Flash")
        self.ota_button.layout.visibility = 'hidden'
        self.refresh_button = Button(
            icon='refresh', tooltip='Refresh File List', layout=Layout(width='15%'))

        if self._dsk:
            self._refresh()

        self.ota_button.on_click(self.on_generate_button_clicked)
        self.flash_button.on_click(self.on_flash_button_clicked)
        self.refresh_button.on_click(self._refresh_platform_files)

        return VBox([
            HBox([widgets.interactive(self.select_platform,
                                      platform_name=self.platform_widget), self.description_widget]),
            HBox([self.flashable_files_widget, self.refresh_button]),
            HBox([self.flash_methods_widget, self.flash_button, self.ota_button])
        ])


def unzip_folder(path_to_zip_file):

    base_name = os.path.basename(path_to_zip_file[:-4])
    hex_path = os.path.join(path_to_zip_file[:-4], base_name+'.hex')

    zip_ref = zipfile.ZipFile(path_to_zip_file, 'r')
    zip_ref.extractall(path_to_zip_file[:-4])
    zip_ref.close()

    return hex_path


def write_key_file(path_to_keyfile):
    example_key_file = ["-----BEGIN EC PRIVATE KEY-----",
                        "MHcCAQEEIHAamBrS46EBrnzzSndN7Gbmc89sj4LLsURIed8K6ynRoAoGCCqGSM49",
                        "AwEHoUQDQgAEFbkmg51LDZLkhpGxi/aqCGK7cZ0r20CVTFz4geb6t89YyNCamxJZ",
                        "cIEyi5PxDhWRvBy4pksO7AOdjFiGj4eGkg==",
                        "-----END EC PRIVATE KEY-----"]

    with open(path_to_keyfile, 'w') as out:
        for line in example_key_file:
            out.write(line)
            out.write('\n')



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


import json
from pandas import DataFrame, Series

from sensiml.datamanager.platformdescription import PlatformDescription
import sensiml.base.utility as utility


class PlatformDescriptions():
    """Base class for a collection of functions"""

    def __init__(self, connection):
        self._connection = connection
        self.build_platform_descriptions()

    def __getitem__(self, index):
        if len(self.platform_list) == 0:
            self.build_platform_descriptions()

        df = self()
        for p in self.platform_list:
            if index not in df.index:
                continue
            if df.loc[index].equals(p().loc[0]):
                return p
        return None

    def refresh(self):
        self.build_platform_descriptions()

    def build_platform_descriptions(self):
        """Populates the platform_list property from the server."""
        self.platform_list = []
        self.platform_dict = {}

        platforms = self.get_platforms()
        for platform in platforms:
            self.platform_dict['{} {}'.format(
                platform.board_name, platform.platform_version)] = platform
            self.platform_list.append(platform)

    def get_platform_by_name(self, name):
        if len(self.platform_list) == 0:
            self.build_platform_descriptions()
        return self.platform_dict.get(name, None)

    def get_platform_by_id(self, id):
        if len(self.platform_list) == 0:
            self.build_platform_descriptions()

        for platform in self.platform_list:
            if platform.id == id:
                return platform
        return None

    def _new_platform_description_from_dict(self, dict):
        """Creates and populates a function from a set of properties.

            Args:
                dict (dict): contains properties of a function

            Returns:
                function
        """
        function = PlatformDescription(self._connection)
        function.initialize_from_dict(dict)
        return function

    def get_platforms(self, function_type=''):
        """Gets all functions as function objects.

            Args:
                function_type (optional[str]): type of function to retrieve

            Returns:
                list of functions
        """
        url = 'platforms/'
        response = self._connection.request('get', url,)
        try:
            response_data, err = utility.check_server_response(response)
        except ValueError:
            print(response)

        platformDescriptions = []
        for platformdesc in response_data:
            platformDescriptions.append(
                self._new_platform_description_from_dict(platformdesc))

        return platformDescriptions

    def __call__(self):
        return self.__str__()

    def __str__(self):
        all_platforms = self.get_platforms()
        if(len(all_platforms) < 0):
            return DataFrame()
        ret = DataFrame(p.__dict__() for p in all_platforms)
        # for plat in all_platforms:
        #     ret = ret.append(plat(), ignore_index=True)
        return ret.sort_values(by='Id').drop(labels=['Id'], axis=1)



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


import os
import json
from sensiml.base import utility


class Function(object):
    """Base class for a transform object"""
    _uuid = ''
    _type = None
    _name = None
    _function_in_file = ''
    _description = ''
    _input_contract = ''
    _subtype = ''
    _has_c_version = False

    def __init__(self, connection):
        self._connection = connection

    @property
    def uuid(self):
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def type(self):
        return self._type

    @property
    def has_c_version(self):
        return self._has_c_version

    @has_c_version.setter
    def has_c_version(self, value):
        self._has_c_version = value

    @type.setter
    def type(self, value):
        self._type = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def input_contract(self):
        return self._input_contract

    @input_contract.setter
    def input_contract(self, value):
        self._input_contract = value

    @property
    def subtype(self):
        return self._subtype

    @subtype.setter
    def subtype(self, value):
        self._subtype = value

    def get_function_info(self):
        """Returns function properties after checking integrity constraints.

            Returns:
                (dict): containing the function's name, type, subtype, and has_c_version

            Raises:
                error if function does not have a name
                error if function does not have a type
        """
        # Sanity Checks
        try:
            assert(self.name is not None)
        except Exception as e:
            print("Function name cannot be None")
            raise(e)

        try:
            assert(self.type is not None)
        except Exception as e:
            print(
                "Function type cannot be None, please set to the correct type (ie. Transform, Optimizer, etc.)")
            raise(e)

        function_info = {'name': self.name,
                         'type': self.type,
                         'subtype': self.subtype,
                         'has_c_version': self.has_c_version,
                         }

        return function_info

    def refresh(self):
        """Calls the REST API and populates the local object properties from the server."""
        url = 'transform/{0}/'.format(self.uuid)
        response = self._connection.request('get', url,)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.name = response_data['name']
            self.type = response_data['type']
            self.subtype = response_data['subtype']
            self.description = response_data['description']
            self.input_contract = response_data['input_contract']
            self.has_c_version = response_data['has_c_version']

    def data(self):
        """Calls the REST API and request a transform file's binary data."""
        url = 'transform/{0}/data/'.format(self.uuid)
        response = self._connection.request('get', url,)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.initialize_from_dict(response_data)

    def initialize_from_dict(self, input_dictionary):
        """Populates a single transform object from a dictionary of properties from the server.

            Args:
                input_dictionary (dict): containing uuid, type, subtype, name, function_in_file, description,
                input_contract, and subtype
        """
        self.uuid = input_dictionary['uuid']
        self.type = input_dictionary['type']
        self.subtype = input_dictionary['subtype']
        self.name = input_dictionary['name']
        self.description = input_dictionary['description']
        self.input_contract = input_dictionary['input_contract']
        self.subtype = input_dictionary['subtype']
        self.has_c_version = input_dictionary['has_c_version']

    def __str__(self):
        input_string = ''
        for input_ in [i for i in self.input_contract if i['name'] not in ['classifiers', 'validation_methods',
                                                                           'number_of_times', 'sample_size']]:
            if 'element_type' in input_:
                input_string += '\n    {name} ({type} of {element_type})'.format(
                    **input_)
            else:
                input_string += '\n    {name} ({type})'.format(**input_)
            if 'options' in input_:
                options_string = ''
                for option in input_['options']:
                    options_string += '{name}, '.format(**option)
                input_string += ' {{options: {0}}}'.format(options_string[:-2])
            if 'description' in input_:
                input_string += ': {description}'.format(**input_)

        return ('NAME: {0}\n'
                'TYPE: {1}\n'
                'SUBTYPE: {2}\n'
                'DESCRIPTION: {3}\n'
                ).format(self.name, self.type, self.subtype,
                         self.description)



from ipywidgets import Layout, Button, VBox, HBox, Dropdown, Checkbox, Label, widgets, HTML
from sensiml.widgets.base_widget import BaseWidget
import qgrid
from pandas import DataFrame
from sensiml.widgets.renderers import WidgetAttributeRenderer


class AutoSenseWidget(BaseWidget):

    def _get_query_list(self):
        q_list = self._dsk.list_queries()
        if q_list is not None:
            return ['']+list(q_list['Name'].values)
        else:
            return ['']

    def _on_button_clicked(self, b):
        self._b_run.disabled = True
        self._b_run.description = 'RUNNING...'
        self._b_run.icon = ''

        try:
            self._run_pipeline_optimizer()
            self._dsk._auto_sense_ran = True
        except:
            self._reset_optimizer_button()

        self._reset_optimizer_button()

        # thread = threading.Thread(target=self._run_pipeline_optimizer)
        # thread.start()

    def _reset_optimizer_button(self):

        self._b_run.disabled = False
        self._b_run.description = 'Optimize Knowledge Pack'
        self._b_run.icon = 'play'

    def _run_pipeline_optimizer(self):
        query_name = self._w_query.value
        segmenter = str(self._w_segment.value)
        seed_choice = str(self._w_seed.value)
        accuracy = self._w_accuracy.value
        sensitivity = self._w_sensitivity.value
        features = self._w_features.value
        classifiers_sram = self._w_classifiers_sram.value
        iterations = self._w_iterations.value
        population_size = self._w_population_size.value
        reset = self._c_reset.value
        allow_unk = self._c_allow_unk.value
        auto_group = self._c_auto_group.value

        if not query_name:
            self.renderer.render("\nERROR: No query name specified!\n")
            return

        if self._dsk.pipeline is None:
            self.renderer.render(
                "\nERROR: Pipeline is not set, use: dsk.pipeline='Pipeline Name'!\n")
            return

        params = {'iterations': iterations,
                  'reset': reset,
                  'iterations': iterations,
                  'population_size': population_size,
                  'allow_unknown': allow_unk,
                  'auto_group': auto_group,
                  'fitness': {'accuracy': accuracy,
                              'sensitivity': sensitivity,
                              'features': features,
                              'classifiers_sram': classifiers_sram}
                  }

        add_windowing = False
        delta = None

        if segmenter == 'Query Segmenter':
            query = self._dsk.project.queries.get_or_create_query(query_name)
            if self._dsk.project.get_segmenters().loc[query.segmenter]['name'] == 'Manual':
                self.renderer.render(
                    "\n\nERROR: This Query only has manually labeled segments associated with it!\n\tSelect a Windowing Segmenter to run this pipeline.\n")
                return

        elif 'Windowing' in segmenter:
            if "(100)" in segmenter:
                delta = 100
            if "(250)" in segmenter:
                delta = 250
            if "(500)" in segmenter:
                delta = 500
            if "(1000)" in segmenter:
                delta = 1000
            if "(2000)" in segmenter:
                delta = 2000
            if "(4000)" in segmenter:
                delta = 4000
            if "(8000)" in segmenter:
                delta = 8000

            add_windowing = True

        else:
            self.renderer.render(
                "\n\nERROR: No segmentation algorithm was selected. Select a Segmenter to run this pipeline.\n")
            return

        self._dsk.project.query_optimize()

        self._dsk.pipeline.reset()
        self._dsk.pipeline.set_input_query(query_name)

        if add_windowing:
            self._dsk.pipeline.add_transform("Windowing",
                                             params={"window_size": delta,
                                                     "delta": delta})

        print("Auto Sense Params",  params)
        self._dsk.pipeline.describe()

        self._clear_auto_sense_results()
        self.results, self.summary = self._dsk.pipeline.auto({'seed': seed_choice,
                                                              'params': params,
                                                              },
                                                             renderer=self.renderer)

        self.renderer.render(
            "\nAutomation Pipeline Completed.")

        self._set_auto_sense_results()

    def _clear_auto_sense_results(self):
        self._w_results.df = DataFrame(
            columns=['Accuracy', 'Sensitivity', 'Model Size (bytes)', 'Features'])

    def _set_auto_sense_results(self):

        self._clear_auto_sense_results()

        if self.summary is not None and isinstance(self.summary, dict) and isinstance(self.summary.get('fitness_summary', None), DataFrame):
            df_results = self.summary['fitness_summary'][[
                'accuracy', 'sensitivity', 'classifiers_sram', 'features']].astype(int).head()
            df_results = df_results.rename(
                index=str, columns={"classifiers_sram": "Model Size"})
            df_results = df_results.rename(
                index=str, columns={"accuracy": "Accuracy"})
            df_results = df_results.rename(
                index=str, columns={"sensitivity": "Sensitivity"})
            df_results = df_results.rename(
                index=str, columns={"features": "Features"})
            self._w_results.df = df_results
        else:
            if isinstance(self.results, dict):
                self.renderer.render(self.results.get("message", ''))

    def _terminate_run(self, b):
        self._dsk.pipeline.stop_pipeline()

    def _select_seed(self, Name):
        if self._dsk and Name:
            self._w_seed_desc.value = self._dsk.seeds.get_seed_by_name(
                Name).description

    def _on_value_change(self, change):
        if self._dsk is None:
            return
        if self._dsk.pipeline and change['new']:
            self._dsk.pipeline.reset()
            self._dsk.pipeline.set_input_query(change['new'])
        else:
            self.renderer.render("No Pipeline Selected.")

    def _refresh(self, b=None):
        if self._dsk:
            self._w_query.options = self._get_query_list()
            self._w_query.values = self._w_query.options[0]
            self._w_seed.options = sorted(self._dsk.seeds.seed_dict.keys())

            if self._dsk.pipeline:
                self._w_pipeline_label.value = "Pipeline Name: " + self._dsk.pipeline.name
                self.results, self.summary = self._dsk.pipeline.get_results(
                    renderer=self.renderer)
                self._set_auto_sense_results()

    def _clear(self):
        self._w_query.options = ['']
        self._w_query.value = ''

    def create_widget(self):

        self._w_pipeline_label = Label(value='Pipeline Name:')

        self._w_query = Dropdown(
            description='Select Query', options=[], layout=Layout(width='85%'))

        self._w_segment = widgets.Dropdown(description='Segmenter',
                                           options=['', 'Query Segmenter', 'Windowing(100)', 'Windowing(250)', 'Windowing(500)', 'Windowing(1000)', 'Windowing(2000)', 'Windowing(4000)', 'Windowing(8000)'])

        self._info_sensitivity = widgets.Button(
            icon="question",
            disabled=True,
            tooltip='Defines the priority of Sensitivity, ie. the ability to determine individual event correctly.\nA higher value corresponds to a higher priority.',
            layout=Layout(width="10%", align_self='flex-end')
        )
        self._info_accuracy = widgets.Button(
            icon="question",
            disabled=True,
            tooltip='Defines the priority of accuracy, ie. the ability to differentiate events correctly.\nA higher value corresponds to a higher priority.',
            layout=Layout(width="10%", align_self='flex-end')
        )
        self._info_classifiers_sram = widgets.Button(
            icon="question",
            disabled=True,
            tooltip='Defines the priority minimize the model size.\nA higher value corresponds to favoring models with smaller sizes i.e. less SRAM usage.',
            layout=Layout(width="10%", align_self='flex-end')
        )
        self._info_features = widgets.Button(
            icon="question",
            disabled=True,
            tooltip='Defines the priority of the number of features, ie. the number of algorithms to differentiate the events correctly.\nLess features typically means less memory and lower latency.\nA higher value corresponds to a higher priority to favor models with fewer features.',
            layout=Layout(width="10%", align_self='flex-end')
        )

        self._info_population = widgets.Button(
            icon="question",
            disabled=True,
            tooltip='Defines how large the inital population is. A higher population means a larger initial search space is.\nA higher population typically produces better results but takes more time.',
            layout=Layout(width="10%", align_self='flex-end')
        )

        self._info_iterations = widgets.Button(
            icon="question",
            disabled=True,
            tooltip='Defines the number of iterations the model will go through.\n At each iteration a new population of models is created by mutating the previous iterations population.\nA higher number of iteration produces better results but takes more time.',
            layout=Layout(width="10%", align_self='flex-end')
        )

        self._info_allow_unk = widgets.Button(
            icon="question",
            disabled=True,
            tooltip='Uses classifiers which will return unknown when the class falls outside of the known predictive space.',
            layout=Layout(width="10%", align_self='flex-end')
        )

        self._info_auto_group = widgets.Button(
            icon="question",
            disabled=True,
            tooltip='Optimize the prediction of the models by putting the classes into the two categories.\n It returns models that predict these two sub-groups.',
            layout=Layout(width="10%", align_self='flex-end')
        )

        self._w_seed = widgets.Dropdown(description='Seed', options=[],)
        self._w_seed_desc = widgets.Textarea(
            description='Description', rows=7, disable=True)
        self._w_intereact = widgets.interactive(
            self._select_seed, Name=self._w_seed)
        self._w_accuracy = widgets.FloatSlider(
            description='Accuracy', value=1.0, min=0.0, max=1.0, step=0.05, )
        self._w_sensitivity = widgets.FloatSlider(
            description='Sensitivity', value=1.0, min=0.0, max=1.0, step=0.05,)
        self._w_features = widgets.FloatSlider(
            description='Features', value=.8, min=0.0, max=1.0, step=0.05,)
        self._w_classifiers_sram = widgets.FloatSlider(
            description='Model Size', value=.8, min=0.0, max=1.0, step=0.05,)
        self._w_iterations = widgets.IntSlider(
            description="Iterations", value=4, min=1, max=15, step=1,)

        self._w_population_size = widgets.IntSlider(
            description="Population", value=25, min=20, max=100, step=1,)
        self._c_reset = Checkbox(description='Reset', value=True)
        self._c_allow_unk = Checkbox(description='Allow Unknown', value=False, layout=Layout(
            width='98%', align_self='flex-end'))
        self._c_auto_group = Checkbox(description='Auto Group Labels', value=False, layout=Layout(
            width='98%', align_self='flex-end'))

        self._w_results = qgrid.show_grid(DataFrame(columns=['Accuracy', 'Sensitivity', 'Model Size (bytes)', 'Features']),
                                          grid_options={'rowHeight': 48,
                                                        'maxVisibleRows': 5,
                                                        'minVisibleRows': 5,
                                                        'editable': False,
                                                        'defaultColumnWidth': 15,
                                                        "forceFitColumns": True,
                                                        'sortable': False,
                                                        'filterable': False}
                                          )

        self._w_results.layout.width = "95%"  # unicode("250px")"
        self._w_results.layout.align_self = 'flex-end'

        self._b_run = Button(icon='play', description='Optimize Knowledge Pack', layout=Layout(
            width='98%', align_self='flex-end'))
        self._b_refresh = Button(icon="refresh", layout=Layout(width='15%'))
        self._b_terminate = Button(icon='stop', description='Terminate', layout=Layout(
            width='25%', align_self='flex-end'))
        self._b_iterate = Button(icon='redo', description='Iterate', layout=Layout(
            width='98%', align_self='flex-end'))

        self._w_query.observe(self._on_value_change, names='value')

        self._b_run.on_click(self._on_button_clicked)
        self._b_refresh.on_click(self._refresh)
        self._b_run.style.button_color = '#4cb243'
        self._b_terminate.on_click(self._terminate_run)

        self._refresh()

        self._widget_render_space = HTML()

        self.renderer = WidgetAttributeRenderer(
            self._widget_render_space, 'value')

        self._widget = VBox([HBox([
            VBox([self._w_pipeline_label,
                  HBox([self._w_query, self._b_refresh]),
                  self._w_segment,
                  self._w_intereact,
                  self._w_seed_desc,
                  HBox(
                      [self._c_allow_unk, self._info_allow_unk]),
                  HBox(
                      [self._c_auto_group, self._info_auto_group]),
                  self._b_run
                  ]),
            VBox([
                HBox([
                    VBox([
                        HBox(
                            [self._w_accuracy, self._info_accuracy],),
                        HBox([self._w_sensitivity,
                              self._info_sensitivity]),
                        HBox(
                            [self._w_features, self._info_features]),
                    ]),
                    VBox(
                        [HBox([self._w_classifiers_sram, self._info_classifiers_sram]),
                         HBox(
                            [self._w_population_size, self._info_population]),
                         HBox(
                            [self._w_iterations, self._info_iterations]),
                         ])
                ]),

                self._w_results,
                self._widget_render_space

            ], layout=Layout(width="66%")),

        ],
            layout=Layout(width='100%')),
        ])

        return self._widget
